"""Arnold-Boffi-Falk elements on quadrilaterals.

Thse elements definitions appear in https://dx.doi.org/10.1137/S0036142903431924
(Arnold, Boffi, Falk, 2005)
"""

import typing

from ..finite_element import CiarletElement
from ..functionals import (IntegralMoment, IntegralOfDivergenceAgainst, ListOfFunctionals,
                           NormalIntegralMoment)
from ..functions import FunctionInput
from ..moments import make_integral_moment_dofs
from ..references import Reference
from ..symbols import x
from .lagrange import Lagrange
from .q import Nedelec


class ArnoldBoffiFalk(CiarletElement):
    """An Arnold-Boffi-Falk element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        assert reference.name == "quadrilateral"
        poly: typing.List[FunctionInput] = []
        poly += [
            (x[0] ** i * x[1] ** j, 0)
            for i in range(order + 3) for j in range(order + 1)]
        poly += [(0, x[0] ** i * x[1] ** j)
                 for i in range(order + 1) for j in range(order + 3)]

        dofs: ListOfFunctionals = make_integral_moment_dofs(
            reference,
            edges=(NormalIntegralMoment, Lagrange, order, {"variant": variant}),
            faces=(IntegralMoment, Nedelec, order, {"variant": variant})
        )

        for i in range(order + 1):
            dofs.append(IntegralOfDivergenceAgainst(
                reference, reference, x[0] ** (order + 1) * x[1] ** i,
                entity=(2, 0), mapping="contravariant"))
            dofs.append(IntegralOfDivergenceAgainst(
                reference, reference, x[0] ** i * x[1] ** (order + 1),
                entity=(2, 0), mapping="contravariant"))

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["Arnold-Boffi-Falk", "ABF"]
    references = ["quadrilateral"]
    min_order = 0
    continuity = "H(div)"



"""Abstract basis function classes and functions."""

from __future__ import annotations

import typing
from abc import abstractmethod

import sympy

from .functions import AnyFunction, FunctionInput, ScalarFunction, SympyFormat, ValuesToSubstitute
from .geometry import PointType
from .references import Reference
from .symbols import AxisVariables, AxisVariablesNotSingle, t


class BasisFunction(AnyFunction):
    """A basis function of a finite element.

    This basis function can be used before the element's basis functions have been computed. When
    the explicit basis function is needed, only then will it be computed.
    """

    def __init__(self, scalar=False, vector=False, matrix=False):
        """Create a basis function.

        Args:
            scalar: Is the function a scalar?
            vector: Is the function a vector?
            matrix: Is the function a matrix?
        """
        super().__init__(scalar=scalar, vector=vector, matrix=matrix)

    @abstractmethod
    def get_function(self) -> AnyFunction:
        """Get the actual basis function.

        Returns:
            The basis function
        """
        pass

    def __add__(self, other: typing.Any) -> AnyFunction:
        """Add.

        Args:
            other: A function to add to this function.

        Returns:
            The sum of two functions
        """
        return self.get_function().__add__(other)

    def __radd__(self, other: typing.Any) -> AnyFunction:
        """Add.

        Args:
            other: A function to add to this function.

        Returns:
            The sum of two functions
        """
        return self.get_function().__radd__(other)

    def __sub__(self, other: typing.Any) -> AnyFunction:
        """Subtract.

        Args:
            other: A function to subtract this function.

        Returns:
            The difference of two functions
        """
        return self.get_function().__sub__(other)

    def __rsub__(self, other: typing.Any) -> AnyFunction:
        """Subtract.

        Args:
            other: A function to subtract this function from.

        Returns:
            The difference of two functions
        """
        return self.get_function().__rsub__(other)

    def __neg__(self) -> AnyFunction:
        """Negate.

        Returns:
            Negated function
        """
        return self.get_function().__neg__()

    def __truediv__(self, other: typing.Any) -> AnyFunction:
        """Divide.

        Args:
            other: A function to divide this function by.

        Returns:
            The ratio of two functions
        """
        return self.get_function().__truediv__(other)

    def __rtruediv__(self, other: typing.Any) -> AnyFunction:
        """Divide.

        Args:
            other: A function to divide by this function.

        Returns:
            The ratio of two functions
        """
        return self.get_function().__rtruediv__(other)

    def __mul__(self, other: typing.Any) -> AnyFunction:
        """Multiply.

        Args:
            other: A function to multiply by this function.

        Returns:
            The product of two functions
        """
        return self.get_function().__mul__(other)

    def __rmul__(self, other: typing.Any) -> AnyFunction:
        """Multiply.

        Args:
            other: A function to multiply by this function.

        Returns:
            The product of two functions
        """
        return self.get_function().__rmul__(other)

    def __matmul__(self, other: typing.Any) -> AnyFunction:
        """Multiply.

        Args:
            other: A function to matrix multiply by this function.

        Returns:
            The product of two matrix functions
        """
        return self.get_function().__matmul__(other)

    def __rmatmul__(self, other: typing.Any) -> AnyFunction:
        """Multiply.

        Args:
            other: A function to matrix multiply by this function.

        Returns:
            The product of two matrix functions
        """
        return self.get_function().__rmatmul__(other)

    def __pow__(self, other: typing.Any) -> AnyFunction:
        """Raise to a power.

        Args:
            other: A power to raise this function to.

        Returns:
            This function to the power of other
        """
        return self.get_function().__pow__(other)

    def as_sympy(self) -> SympyFormat:
        """Convert to a sympy expression.

        Returns:
            This function as a Sympy expression
        """
        return self.get_function().as_sympy()

    def as_tex(self) -> str:
        """Convert to a TeX expression.

        Returns:
            A TeX representation of this function
        """
        return self.get_function().as_tex()

    def diff(self, variable: sympy.core.symbol.Symbol) -> AnyFunction:
        """Differentiate the function.

        Args:
            variable: The variable to differentiate with respect to

        Returns:
            The derivative
        """
        return self.get_function().diff(variable)

    def directional_derivative(self, direction: PointType) -> AnyFunction:
        """Compute a directional derivative.

        Args:
            direction: The direction of the derivative

        Returns:
            The derivative
        """
        return self.get_function().directional_derivative(direction)

    def jacobian_component(self, component: typing.Tuple[int, int]) -> AnyFunction:
        """Compute a component of the jacobian.

        Args:
            component: The component to compute

        Returns:
            Component of the Jacobian
        """
        return self.get_function().jacobian_component(component)

    def jacobian(self, dim: int) -> AnyFunction:
        """Compute the jacobian.

        Args:
            dim: The dimension of the Jacobian

        Returns:
            The Jabobian
        """
        return self.get_function().jacobian(dim)

    def dot(self, other_in: FunctionInput) -> AnyFunction:
        """Compute the dot product with another function.

        Args:
            other_in: The function to dot with

        Return:
            The dot product
        """
        return self.get_function().dot(other_in)

    def cross(self, other_in: FunctionInput) -> AnyFunction:
        """Compute the cross product with another function.

        Args:
            other_in: The function to cross with

        Return:
            The cross product
        """
        return self.get_function().cross(other_in)

    def div(self) -> AnyFunction:
        """Compute the divergence of the function.

        Returns:
            The divergence
        """
        return self.get_function().div()

    def grad(self, dim: int) -> AnyFunction:
        """Compute the gradient of the function.

        Returns:
            The gradient
        """
        return self.get_function().grad(dim)

    def curl(self) -> AnyFunction:
        """Compute the curl of the function.

        Returns:
            The curl
        """
        return self.get_function().curl()

    def norm(self) -> ScalarFunction:
        """Compute the norm of the function.

        Returns:
            The norm
        """
        raise self.get_function().norm()

    def integral(self, domain: Reference, vars: AxisVariablesNotSingle = t) -> AnyFunction:
        """Compute the integral of the function.

        Args:
            domain: The domain to integrate over
            vars: The variables to integrate over

        Returns:
            The integral
        """
        return self.get_function().integral(domain, vars)

    def subs(self, vars: AxisVariables, values: ValuesToSubstitute) -> BasisFunction:
        """Substitute values into the function.

        Args:
            vars: The variable(s) to substitute
            values: The value(s) to substitute

        Returns:
            Substituted function
        """
        return SubbedBasisFunction(self, vars, values)

    def __getitem__(self, key) -> AnyFunction:
        """Forward all other function calls to symbolic function."""
        return self.get_function().__getitem__(key)

    def __len__(self) -> int:
        """Get length.

        Return:
            The length
        """
        return self.get_function().__len__()

    def det(self) -> ScalarFunction:
        """Compute the determinant.

        Returns:
            The determinant
        """
        if not self.is_matrix:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute 'det'")
        return self.get_function().det()

    def transpose(self) -> ScalarFunction:
        """Compute the transpose.

        Returns:
            The transpose
        """
        if not self.is_matrix:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute 'transpose'")
        return self.get_function().transpose()

    def with_floats(self) -> AnyFunction:
        """Return a version the function with floats as coefficients.

        Returns:
            The function with floats as coefficients
        """
        return self.get_function().with_floats()


class SubbedBasisFunction(BasisFunction):
    """A basis function following a substitution."""

    def __init__(
        self, f: BasisFunction, vars: AxisVariables, values: ValuesToSubstitute
    ):
        """Create a basis function following a substitution.

        Args:
            f: The basis function
            vars: The variables that have been substituted
            values: The values to substitute
        """
        super().__init__(scalar=f.is_scalar, vector=f.is_vector, matrix=f.is_matrix)
        self.f = f
        self._vars = vars
        self._values = values

    def get_function(self) -> AnyFunction:
        """Return the symbolic function.

        Returns:
            Symbolic function
        """
        return self.f.get_function().subs(self._vars, self._values)

    @property
    def shape(self) -> typing.Tuple[int, ...]:
        """Get the value shape of the function.

        Returns:
            shape
        """
        return self.f.get_function().shape

    def plot_values(
        self, reference: Reference, img: typing.Any,
        value_scale: sympy.core.expr.Expr = sympy.Integer(1), n: int = 6
    ):
        """Plot the function's values.

        Args:
            reference: The domain to plot the function on
            img: The image to plot the values on
            value_scale: The factor to scale values by
            n: The number of plotting points
        """
        self.f.plot_values(reference, img, value_scale, n)



"""Bernardi-Raugel elements on simplices.

This element's definition appears in https://doi.org/10.2307/2007793
(Bernardi and Raugel, 1985)
"""

import typing

import sympy

from ..finite_element import CiarletElement
from ..functionals import (DivergenceIntegralMoment, DotPointEvaluation, ListOfFunctionals,
                           NormalIntegralMoment)
from ..functions import FunctionInput
from ..moments import make_integral_moment_dofs
from ..polynomials import polynomial_set_vector
from ..references import Reference
from ..symbols import x
from .lagrange import Lagrange


class BernardiRaugel(CiarletElement):
    """Bernardi-Raugel Hdiv finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_vector(reference.tdim, reference.tdim, order)

        p = Lagrange(reference, 1, variant="equispaced")

        for i in range(reference.sub_entity_count(reference.tdim - 1)):
            sub_e = reference.sub_entity(reference.tdim - 1, i)
            bubble = sympy.Integer(1)
            for j in reference.sub_entities(reference.tdim - 1)[i]:
                bubble *= p.get_basis_function(j)
            poly.append(tuple(bubble * j for j in sub_e.normal()))

        dofs: ListOfFunctionals = []

        for n in range(reference.sub_entity_count(reference.tdim - 1)):
            facet = reference.sub_entity(reference.tdim - 1, n)
            for v in facet.vertices:
                dofs.append(DotPointEvaluation(
                    reference, v, tuple(i * facet.jacobian() for i in facet.normal()),
                    entity=(reference.tdim - 1, n), mapping="contravariant"))

        dofs += make_integral_moment_dofs(
            reference,
            facets=(NormalIntegralMoment, Lagrange, 0, "contravariant",
                    {"variant": "equispaced"}),
        )

        if order > 1:
            assert order == 2 and reference.name == "tetrahedron"

            for i in range(reference.tdim):
                bf = p.get_basis_functions()
                poly.append(tuple(
                    bf[0] * bf[1] * bf[2] * bf[3] if j == i else 0 for j in range(reference.tdim)))

            for e_n, edge in enumerate(reference.edges):
                v1 = reference.vertices[edge[0]]
                v2 = reference.vertices[edge[1]]
                midpoint = tuple(sympy.Rational(i + j, 2) for i, j in zip(v1, v2))
                d = tuple(j - i for i, j in zip(v1, v2))
                dofs.append(DotPointEvaluation(reference, midpoint, d, entity=(1, e_n),
                                               mapping="contravariant"))
            for f_n in range(reference.sub_entity_count(2)):
                face = reference.sub_entity(2, f_n)
                normal = tuple(i * face.jacobian() for i in face.normal())
                for e_n in range(3):
                    edge_entity = face.sub_entity(1, e_n)
                    midpoint = tuple(
                        sympy.Rational(i + j, 2) for i, j in zip(*edge_entity.vertices))
                    dofs.append(DotPointEvaluation(
                        reference, midpoint, normal, entity=(2, f_n), mapping="contravariant"))

            p = Lagrange(reference, 0, variant="equispaced")

            for i in range(3):
                dofs.append(DivergenceIntegralMoment(
                    reference, reference, x[i], p.dofs[0], entity=(3, 0),
                    mapping="contravariant"
                ))

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)

    names = ["Bernardi-Raugel"]
    references = ["triangle", "tetrahedron"]
    min_order = 1
    max_order = {"triangle": 1, "tetrahedron": 2}
    continuity = "H(div)"



"""Test Bernstein elements."""

import pytest

import symfem


@pytest.mark.parametrize("celltype, degree", [(c, i) for c, n in [
    ("interval", 4), ("triangle", 3), ("tetrahedron", 2)
] for i in range(n)])
def test_bernstein(celltype, degree):
    b = symfem.create_element(celltype, "Bernstein", degree)
    poly = symfem.elements.bernstein.bernstein_polynomials(degree, b.reference.tdim)

    for f in b.get_basis_functions():
        for p in poly:
            if f == p:
                poly.remove(p)
                break
        else:
            raise ValueError(f"{f} is not a Bernstein polynomial.")

    assert len(poly) == 0



"""Lagrange elements on simplices."""

import typing
from itertools import product

import sympy

from ..finite_element import CiarletElement
from ..functionals import DotPointEvaluation, ListOfFunctionals, PointEvaluation
from ..functions import FunctionInput
from ..polynomials import polynomial_set_1d, polynomial_set_vector
from ..quadrature import get_quadrature
from ..references import Reference


class Lagrange(CiarletElement):
    """Lagrange finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        dofs: ListOfFunctionals = []
        if order == 0:
            dofs = [
                PointEvaluation(
                    reference, tuple(
                        sympy.Rational(1, reference.tdim + 1)
                        for i in range(reference.tdim)
                    ),
                    entity=(reference.tdim, 0)
                )
            ]
        else:
            points, _ = get_quadrature(variant, order + 1)

            for v_n, v in enumerate(reference.reference_vertices):
                dofs.append(PointEvaluation(reference, v, entity=(0, v_n)))
            for edim in range(1, 4):
                for e_n in range(reference.sub_entity_count(edim)):
                    entity = reference.sub_entity(edim, e_n)
                    for i in product(range(1, order), repeat=edim):
                        if sum(i) < order:
                            point = entity.get_point([sympy.Rational(j, order) for j in i[::-1]])
                            dofs.append(PointEvaluation(reference, point, entity=(edim, e_n)))

        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_1d(reference.tdim, order)
        super().__init__(reference, order, poly, dofs, reference.tdim, 1)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["Lagrange", "P"]
    references = ["interval", "triangle", "tetrahedron"]
    min_order = 0
    continuity = "C0"


class VectorLagrange(CiarletElement):
    """Vector Lagrange finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        scalar_space = Lagrange(reference, order, variant)
        dofs: ListOfFunctionals = []
        poly: typing.List[FunctionInput] = []
        if reference.tdim == 1:
            for p in scalar_space.dofs:
                dofs.append(PointEvaluation(reference, p.dof_point(), entity=p.entity))

            poly += polynomial_set_1d(reference.tdim, order)
        else:
            directions = [
                tuple(1 if i == j else 0 for j in range(reference.tdim))
                for i in range(reference.tdim)
            ]
            for p in scalar_space.dofs:
                for d in directions:
                    dofs.append(DotPointEvaluation(reference, p.dof_point(), d, entity=p.entity))

            poly += polynomial_set_vector(reference.tdim, reference.tdim, order)
        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["vector Lagrange", "vP"]
    references = ["interval", "triangle", "tetrahedron"]
    min_order = 0
    continuity = "C0"


class MatrixLagrange(CiarletElement):
    """Matrix Lagrange finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        scalar_space = Lagrange(reference, order, variant)
        dofs: ListOfFunctionals = []
        if reference.tdim == 1:
            directions: typing.List[typing.Tuple[int, ...]] = [(1, )]
        else:
            directions = [
                tuple(1 if i == j else 0 for j in range(reference.tdim ** 2))
                for i in range(reference.tdim ** 2)
            ]
        for p in scalar_space.dofs:
            for d in directions:
                dofs.append(DotPointEvaluation(reference, p.dof_point(), d, entity=p.entity))

        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_vector(reference.tdim, reference.tdim ** 2, order)

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim ** 2,
                         (reference.tdim, reference.tdim))
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["matrix Lagrange"]
    references = ["triangle", "tetrahedron"]
    min_order = 0
    continuity = "L2"


class SymmetricMatrixLagrange(CiarletElement):
    """Symmetric matrix Lagrange finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        poly: typing.List[FunctionInput] = []
        dofs: ListOfFunctionals = []
        scalar_space = Lagrange(reference, order, variant)
        if reference.tdim == 1:
            poly += polynomial_set_1d(1, order)
            for p in scalar_space.dofs:
                dofs.append(PointEvaluation(reference, p.dof_point(), entity=p.entity))
            super().__init__(
                reference, order, poly, dofs,
                reference.tdim, reference.tdim ** 2,
            )
        else:
            directions: typing.List[typing.Tuple[typing.Tuple[int, ...], ...]] = []
            if reference.tdim == 2:
                poly += [((a[0], a[1]), (a[1], a[2])) for a in polynomial_set_vector(2, 3, order)]
                directions = [((1, 0), (0, 0)), ((0, 1), (0, 0)),
                              ((0, 0), (0, 1))]
            else:
                assert reference.tdim == 3
                poly += [((a[0], a[1], a[2]),
                          (a[1], a[3], a[4]),
                          (a[2], a[4], a[5])) for a in polynomial_set_vector(3, 6, order)]
                directions = [((1, 0, 0), (0, 0, 0), (0, 0, 0)),
                              ((0, 1, 0), (0, 0, 0), (0, 0, 0)),
                              ((0, 0, 1), (0, 0, 0), (0, 0, 0)),
                              ((0, 0, 0), (0, 1, 0), (0, 0, 0)),
                              ((0, 0, 0), (0, 0, 1), (0, 0, 0)),
                              ((0, 0, 0), (0, 0, 0), (0, 0, 1))]

            for p in scalar_space.dofs:
                for d in directions:
                    dofs.append(DotPointEvaluation(reference, p.dof_point(), d, entity=p.entity))

            super().__init__(
                reference, order, poly, dofs,
                reference.tdim, reference.tdim ** 2, (reference.tdim, reference.tdim),
            )
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["symmetric matrix Lagrange"]
    references = ["triangle", "tetrahedron"]
    min_order = 0
    continuity = "L2"



"""Symbols."""

import typing

import sympy

x = (sympy.Symbol("x"), sympy.Symbol("y"), sympy.Symbol("z"))
t = (sympy.Symbol("t0"), sympy.Symbol("t1"), sympy.Symbol("t2"))

AxisVariablesNotSingle = typing.Union[
    typing.Tuple[sympy.core.symbol.Symbol, ...],
    typing.List[sympy.core.symbol.Symbol]]
AxisVariables = typing.Union[AxisVariablesNotSingle, sympy.core.symbol.Symbol]



"""Utility functions for testing."""

import typing

test_elements: typing.Dict[str, typing.Dict[
    str,
    typing.List[typing.Tuple[typing.Dict[str, typing.Any], typing.Iterable]]
]] = {
    "interval": {
        "P": [({"variant": "equispaced"}, range(6)), ({"variant": "lobatto"}, range(4))],
        "vP": [({"variant": "equispaced"}, range(6)), ({"variant": "lobatto"}, range(3)),
               ({"variant": "radau"}, range(3)), ({"variant": "legendre"}, range(3))],
        "dPc": [({"variant": "equispaced"}, range(6)), ({"variant": "lobatto"}, range(4))],
        "bubble": [({"variant": "equispaced"}, range(2, 6)),
                   ({"variant": "lobatto"}, range(2, 3))],
        "serendipity": [({"variant": "equispaced"}, range(1, 6)),
                        ({"variant": "lobatto"}, range(1, 3))],
        "Hermite": [({}, [3])],
        "Bernstein": [({}, range(1, 4))],
        "Taylor": [({}, range(0, 5))],
        "Wu-Xu": [({}, [2])],
        "MWX": [({}, [1])],
    },
    "triangle": {
        "P": [({"variant": "equispaced"}, range(5)), ({"variant": "lobatto"}, range(3))],
        "vP": [({"variant": "equispaced"}, range(5)), ({"variant": "lobatto"}, range(3))],
        "matrix Lagrange": [({"variant": "equispaced"}, range(3)),
                            ({"variant": "lobatto"}, range(3))],
        "symmetric matrix Lagrange": [
            ({"variant": "equispaced"}, range(3)), ({"variant": "lobatto"}, range(3))],
        "bubble": [({"variant": "equispaced"}, range(3, 5)), ({"variant": "lobatto"}, range(3, 3))],
        "bubble enriched Lagrange": [({"variant": "equispaced"}, range(1, 3)),
                                     ({"variant": "lobatto"}, range(1, 3))],
        "bubble enriched vector Lagrange":
            [({"variant": "equispaced"}, range(1, 3)), ({"variant": "lobatto"}, range(1, 3))],
        "CR": [({"variant": "equispaced"}, [1, 3, 5])],
        "conforming CR": [({}, range(1, 6))],
        "HHJ": [({"variant": "equispaced"}, range(3))],
        "Bell": [({"variant": "equispaced"}, [5])],
        "Morley": [({}, [2])],
        "MWX": [({}, [1, 2])],
        "Regge": [({"variant": "point"}, range(4)), ({"variant": "integral"}, range(3))],
        "AW": [({"variant": "equispaced"}, range(3, 5))],
        "nonconforming AW": [({"variant": "equispaced"}, [2])],
        "Nedelec1": [({"variant": "equispaced"}, range(1, 4))],
        "Nedelec2": [({"variant": "equispaced"}, range(1, 4))],
        "RT": [({"variant": "equispaced"}, range(1, 4))],
        "BDFM": [({"variant": "equispaced"}, range(1, 4))],
        "Argyris": [({}, [5])],
        "MTW": [({"variant": "equispaced"}, [3])],
        "KMV": [({}, [1, 3])],
        "Hermite": [({}, [3])],
        "BDM": [({"variant": "equispaced"}, range(1, 4))],
        "Bernstein": [({}, range(1, 4))],
        "HCT": [({}, [3])],
        "rHCT": [({}, [3])],
        "FS": [({}, [2])],
        "Taylor": [({}, range(0, 5))],
        "Bernardi-Raugel": [({}, [1])],
        "Guzman-Neilan": [({}, [1])],
        "Wu-Xu": [({}, [3])],
        "transition": [({"edge_orders": [1, 1, 1], "variant": "equispaced"}, range(1, 6)),
                       ({"edge_orders": [2, 2, 1], "variant": "equispaced"}, range(1, 6)),
                       ({"edge_orders": [3, 3, 1], "variant": "equispaced"}, range(1, 6)),
                       ({"edge_orders": [3, 3, 2], "variant": "equispaced"}, range(1, 6)),
                       ({"edge_orders": [2, 2, 1], "variant": "lobatto"}, range(1, 3))],
        "P1-iso-P2": [({}, [1])],
    },
    "tetrahedron": {
        "P": [({"variant": "equispaced"}, range(3)), ({"variant": "lobatto"}, range(3))],
        "vP": [({"variant": "equispaced"}, range(3)), ({"variant": "lobatto"}, range(3))],
        "matrix Lagrange": [({"variant": "equispaced"}, range(2)),
                            ({"variant": "lobatto"}, range(3))],
        "symmetric matrix Lagrange": [({"variant": "equispaced"}, range(2)),
                                      ({"variant": "lobatto"}, range(3))],
        "bubble": [({"variant": "equispaced"}, [4]), ({"variant": "lobatto"}, [4])],
        "CR": [({"variant": "equispaced"}, [1])],
        "Regge": [({"variant": "point"}, range(3)), ({"variant": "integral"}, range(2))],
        "Nedelec1": [({"variant": "equispaced"}, range(1, 3))],
        "Nedelec2": [({"variant": "equispaced"}, range(1, 3))],
        "RT": [({"variant": "equispaced"}, range(1, 3))],
        "BDFM": [({"variant": "equispaced"}, range(1, 3))],
        "MTW": [({"variant": "equispaced"}, [3])],
        "MWX": [({}, [1, 2, 3])],
        "KMV": [({}, [1])],
        "Hermite": [({}, [3])],
        "BDM": [({"variant": "equispaced"}, range(1, 3))],
        "Bernstein": [({}, range(1, 3))],
        "Taylor": [({}, range(0, 5))],
        "Bernardi-Raugel": [({}, [1, 2])],
        "Guzman-Neilan": [({}, [1, 2])],
        "Wu-Xu": [({}, [4])],
        "transition": [({"edge_orders": [1, 1, 1, 1, 1, 1], "face_orders": [1, 1, 1, 1],
                         "variant": "equispaced"}, range(1, 5)),
                       ({"edge_orders": [1, 1, 1, 1, 1, 1], "face_orders": [3, 3, 1, 1],
                         "variant": "equispaced"}, range(1, 5)),
                       ({"edge_orders": [1, 1, 1, 1, 1, 1], "face_orders": [3, 3, 4, 2],
                         "variant": "equispaced"}, range(1, 5)),
                       ({"edge_orders": [1, 1, 1, 1, 1, 1], "face_orders": [4, 4, 1, 1],
                         "variant": "equispaced"}, range(1, 5)),
                       ({"edge_orders": [2, 2, 2, 2, 2, 2], "face_orders": [1, 1, 1, 1],
                         "variant": "equispaced"}, range(1, 5)),
                       ({"edge_orders": [3, 2, 1, 2, 1, 1], "face_orders": [1, 1, 1, 1],
                         "variant": "equispaced"}, range(1, 5)),
                       ({"edge_orders": [1, 1, 3, 1, 3, 1], "face_orders": [1, 1, 1, 1],
                         "variant": "equispaced"}, range(1, 5)),
                       ({"edge_orders": [1, 1, 2, 1, 2, 5], "face_orders": [3, 3, 2, 5],
                         "variant": "equispaced"}, range(1, 5))],
    },
    "quadrilateral": {
        "bubble": [({"variant": "equispaced"}, range(2, 4)), ({"variant": "lobatto"}, range(2, 4))],
        "Q": [({"variant": "equispaced"}, range(4)), ({"variant": "lobatto"}, range(3))],
        "vQ": [({"variant": "equispaced"}, range(4)), ({"variant": "lobatto"}, range(3))],
        "dPc": [({"variant": "equispaced"}, range(4)), ({"variant": "lobatto"}, range(3))],
        "vector dPc": [({"variant": "equispaced"}, range(4)), ({"variant": "lobatto"}, range(3))],
        "serendipity": [({"variant": "equispaced"}, range(1, 4)),
                        ({"variant": "lobatto"}, range(1, 3))],
        "direct serendipity": [({}, range(1, 7))],
        "Scurl": [({"variant": "equispaced"}, range(1, 4)), ({"variant": "lobatto"}, range(1, 3))],
        "Sdiv": [({"variant": "equispaced"}, range(1, 4)), ({"variant": "lobatto"}, range(1, 3))],
        "Qcurl": [({"variant": "equispaced"}, range(1, 4))],
        "Qdiv": [({"variant": "equispaced"}, range(1, 4))],
        "BFS": [({}, [3])],
        "BDFM": [({"variant": "equispaced"}, range(1, 4))],
        "TScurl": [({"variant": "equispaced"}, range(1, 5))],
        "TSdiv": [({"variant": "equispaced"}, range(1, 5))],
        "TNT": [({"variant": "equispaced"}, range(1, 4))],
        "TNTcurl": [({"variant": "equispaced"}, range(1, 4))],
        "TNTdiv": [({"variant": "equispaced"}, range(1, 4))],
        "Regge": [({"variant": "integral"}, range(3))],
        "ABF": [({}, range(3))],
        "AC": [({}, range(4))],
        "Rannacher-Turek": [({}, [1])],
        "P1-iso-P2": [({}, [1])],
        "HZ": [({}, [2, 3, 4])],
    },
    "hexahedron": {
        "bubble": [({"variant": "equispaced"}, range(2, 4)), ({"variant": "lobatto"}, range(2, 4))],
        "Q": [({"variant": "equispaced"}, range(3)), ({"variant": "lobatto"}, range(3))],
        "vQ": [({"variant": "equispaced"}, range(3)), ({"variant": "lobatto"}, range(3))],
        "dPc": [({"variant": "equispaced"}, range(3)), ({"variant": "lobatto"}, range(3))],
        "vector dPc": [({"variant": "equispaced"}, range(3)), ({"variant": "lobatto"}, range(3))],
        "serendipity": [({"variant": "equispaced"}, range(1, 3)),
                        ({"variant": "lobatto"}, range(1, 3))],
        "Scurl": [({"variant": "equispaced"}, range(1, 3)), ({"variant": "lobatto"}, range(1, 3))],
        "Sdiv": [({"variant": "equispaced"}, range(1, 3)), ({"variant": "lobatto"}, range(1, 3))],
        "Qcurl": [({"variant": "equispaced"}, range(1, 3))],
        "Qdiv": [({"variant": "equispaced"}, range(1, 3))],
        "BDFM": [({"variant": "equispaced"}, range(1, 3))],
        "BDDF": [({"variant": "equispaced"}, range(1, 3))],
        "TScurl": [({"variant": "equispaced"}, range(1, 4))],
        "TSdiv": [({"variant": "equispaced"}, range(1, 4))],
        "TNT": [({"variant": "equispaced"}, range(1, 2))],
        "TNTcurl": [({"variant": "equispaced"}, range(1, 2))],
        "TNTdiv": [({"variant": "equispaced"}, range(1, 2))],
        "Regge": [({"variant": "integral"}, range(2))],
        "Rannacher-Turek": [({}, [1])],
    },
    "prism": {
        "Lagrange": [({"variant": "equispaced"}, range(4))],
        "Nedelec": [({"variant": "equispaced"}, range(1, 3))],
    },
    "pyramid": {
        "Lagrange": [({"variant": "equispaced"}, range(4))],
    }
}



"""Nedelec elements on simplices.

These elements' definitions appear in https://doi.org/10.1007/BF01396415
(Nedelec, 1980) and https://doi.org/10.1007/BF01389668 (Nedelec, 1986)
"""

import typing

from ..finite_element import CiarletElement
from ..functionals import IntegralMoment, ListOfFunctionals, TangentIntegralMoment
from ..functions import FunctionInput
from ..moments import make_integral_moment_dofs
from ..polynomials import Hcurl_polynomials, polynomial_set_vector
from ..references import Reference
from .lagrange import Lagrange, VectorLagrange
from .rt import RaviartThomas


class NedelecFirstKind(CiarletElement):
    """Nedelec first kind Hcurl finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_vector(reference.tdim, reference.tdim, order - 1)
        poly += Hcurl_polynomials(reference.tdim, reference.tdim, order)
        dofs: ListOfFunctionals = make_integral_moment_dofs(
            reference,
            edges=(TangentIntegralMoment, Lagrange, order - 1,
                   {"variant": variant}),
            faces=(IntegralMoment, VectorLagrange, order - 2, "covariant",
                   {"variant": variant}),
            volumes=(IntegralMoment, VectorLagrange, order - 3, "covariant",
                     {"variant": variant}),
        )

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["Nedelec", "Nedelec1", "N1curl"]
    references = ["triangle", "tetrahedron"]
    min_order = 1
    continuity = "H(curl)"


class NedelecSecondKind(CiarletElement):
    """Nedelec second kind Hcurl finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_vector(reference.tdim, reference.tdim, order)

        dofs: ListOfFunctionals = make_integral_moment_dofs(
            reference,
            edges=(TangentIntegralMoment, Lagrange, order, {"variant": variant}),
            faces=(IntegralMoment, RaviartThomas, order - 1, "covariant", {"variant": variant}),
            volumes=(IntegralMoment, RaviartThomas, order - 2, "covariant", {"variant": variant}),
        )

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["Nedelec2", "N2curl"]
    references = ["triangle", "tetrahedron"]
    min_order = 1
    continuity = "H(curl)"



"""Arbogast-Correa elements on quadrilaterals.

This element's definition appears in https://dx.doi.org/10.1137/15M1013705
(Arbogast, Correa, 2016)
"""

import typing

from ..finite_element import CiarletElement
from ..functionals import IntegralAgainst, ListOfFunctionals, NormalIntegralMoment
from ..functions import FunctionInput
from ..moments import make_integral_moment_dofs
from ..polynomials import Hdiv_serendipity, polynomial_set_vector
from ..references import Reference
from ..symbols import x
from .dpc import DPC


class AC(CiarletElement):
    """Arbogast-Correa Hdiv finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_vector(reference.tdim, reference.tdim, order)
        if order == 0:
            poly += [(x[0], 0), (0, x[1])]
        else:
            poly += Hdiv_serendipity(reference.tdim, reference.tdim, order)
            poly += [(x[0] ** (i + 1) * x[1] ** (order - i), x[0] ** i * x[1] ** (1 + order - i))
                     for i in range(order + 1)]

        dofs: ListOfFunctionals = make_integral_moment_dofs(
            reference,
            facets=(NormalIntegralMoment, DPC, order, {"variant": variant}),
        )

        for i in range(order + 1):
            for j in range(order + 1 - i):
                if i + j > 0:
                    f = (i * x[0] ** (i - 1) * x[1] ** j, j * x[0] ** i * x[1] ** (j - 1))
                    dofs.append(IntegralAgainst(reference, reference, f, entity=(2, 0),
                                                mapping="contravariant"))

        for i in range(1, order - 1):
            for j in range(1, order - i):
                f = (
                    x[0] ** i * (1 - x[0]) * x[1] ** (j - 1) * (j * (1 - x[1]) - x[1]),
                    -x[0] ** (i - 1) * (i * (1 - x[0]) - x[0]) * x[1] ** j * (1 - x[1])
                )
                dofs.append(IntegralAgainst(reference, reference, f, entity=(2, 0),
                                            mapping="contravariant"))

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["Arbogast-Correa", "AC", "AC full", "Arbogast-Correa full"]
    references = ["quadrilateral"]
    min_order = 0
    continuity = "H(div)"



"""Bernstein elements on simplices.

This element's definition appears in
https://doi.org/10.1007/s00211-010-0327-2 (Kirby, 2011) and
https://doi.org/10.1137/11082539X (Ainsworth, Andriamaro, Davydov, 2011)
"""

import typing

import sympy

from ..finite_element import CiarletElement
from ..functionals import BaseFunctional, ListOfFunctionals, PointEvaluation
from ..functions import AnyFunction, FunctionInput
from ..geometry import PointType
from ..polynomials import orthogonal_basis, polynomial_set_1d
from ..references import Reference
from ..symbols import AxisVariablesNotSingle, t, x


def single_choose(n: int, k: int) -> sympy.core.expr.Expr:
    """Calculate choose function of a set of powers.

    Args:
        n: Number of items
        k: Number to select

    Returns:
        Number of ways to pick k items from n items (ie n choose k)
    """
    out = sympy.Integer(1)
    for i in range(k + 1, n + 1):
        out *= i
    for i in range(1, n - k + 1):
        out /= i
    return out


def choose(n: int, powers: typing.List[int]) -> sympy.core.expr.Expr:
    """Calculate choose function of a set of powers.

    Args:
        n: Number of items
        k: Numbers to select

    Returns:
        A multichoose function
    """
    out = sympy.Integer(1)
    for p in powers:
        out *= single_choose(n, p)
        n -= p
    return out


def bernstein_polynomials(
    n: int, d: int, vars: AxisVariablesNotSingle = x
) -> typing.List[sympy.core.expr.Expr]:
    """Return a list of Bernstein polynomials.

    Args:
        n: The polynomial order
        d: The topological dimension
        vars: The variables to use

    Returns:
        Bernstein polynomials
    """
    poly = []
    if d == 1:
        lambdas = [1 - vars[0], vars[0]]
        powers = [[n - i, i] for i in range(n + 1)]
    elif d == 2:
        lambdas = [1 - vars[0] - vars[1], vars[0], vars[1]]
        powers = [[n - i - j, j, i]
                  for i in range(n + 1)
                  for j in range(n + 1 - i)]
    elif d == 3:
        lambdas = [1 - vars[0] - vars[1] - vars[2], vars[0], vars[1], vars[2]]
        powers = [[n - i - j - k, k, j, i]
                  for i in range(n + 1)
                  for j in range(n + 1 - i)
                  for k in range(n + 1 - i - j)]

    for p in powers:
        f = choose(n, p)
        for a, b in zip(lambdas, p):
            f *= a ** b
        poly.append(f)

    return poly


class BernsteinFunctional(BaseFunctional):
    """Functional for a Bernstein element."""

    def __init__(self, reference: Reference, integral_domain: Reference, index: int,
                 degree: int, entity: typing.Tuple[int, int]):
        """Create the functional.

        Args:
            reference: The reference element
            integral_domain: The subentity to integrate over
            index: The index of the bernstein polynomial
            degree: The polynomial degree
            entity: The entity this functional is associated with
        """
        super().__init__(reference, entity, "identity")
        orth = [
            o / sympy.sqrt((o * o).integral(integral_domain))
            for o in orthogonal_basis(integral_domain.name, degree, 0, t[:integral_domain.tdim])[0]
        ]
        self.ref = integral_domain
        self.index = index
        self.degree = degree

        bern = bernstein_polynomials(degree, integral_domain.tdim, t)
        mat = sympy.Matrix(
            [[(o * b).integral(integral_domain) for b in bern] for o in orth])
        minv = mat.inv()
        alpha = minv.row(index)

        self.moment = sum(i * j for i, j in zip(alpha, orth))

    def dof_point(self) -> PointType:
        """Get the location of the DOF in the cell.

        Returns:
            Location of the DOF
        """
        return self.ref.sub_entity(*self.entity).midpoint()

    def _eval_symbolic(self, function: AnyFunction) -> AnyFunction:
        """Apply the functional to a function.

        Args:
            function: The function

        Returns:
            Evaluation of the functional
        """
        point = [i for i in self.ref.origin]
        for i, a in enumerate(zip(*self.ref.axes)):
            for j, k in zip(a, t):
                point[i] += j * k

        integrand = function.subs(x, point) * self.moment
        return integrand.integral(self.ref)

    def get_tex(self) -> typing.Tuple[str, typing.List[str]]:
        """Get a representation of the functional as TeX, and list of terms involved.

        Returns:
            TeX representation
        """
        if self.reference.tdim == self.ref.tdim:
            return f"v\\mapsto c_{{{self.index}}}", [
                "\\(v=\\sum_ic_iB_i\\)",
                f"\\(B_1\\) to \\(B_n\\) "
                f"are the degree {self.degree} Bernstein polynomials on the cell"
            ]
        else:
            e = self.entity_tex()
            return f"v\\mapsto c^{{{e}}}_{{{self.index}}}", [
                f"\\(v=\\sum_ic^{{{e}}}_iB^{{{e}}}_i\\)",
                f"\\(B^{{{e}}}_1\\) to \\(B^{{{e}}}_n\\) "
                f"are the degree {self.degree} Bernstein polynomials on \\({e}\\)",
                self.entity_definition()
            ]


class Bernstein(CiarletElement):
    """Bernstein finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_1d(reference.tdim, order)

        dofs: ListOfFunctionals = []
        if order == 0:
            dofs = [
                PointEvaluation(reference, reference.midpoint(), (reference.tdim, 0))]
        else:
            def index(x: int, y: int = 0, z: int = 0) -> int:
                """Compute the 1D index."""
                return (
                    z * (z ** 2 - 3 * z * order - 6 * z + 3 * order ** 2 + 12 * order + 11) // 6
                    + y * (2 * (order - z) + 3 - y) // 2
                    + x
                )

            for vn, v in enumerate(reference.vertices):
                dofs.append(PointEvaluation(reference, v, (0, vn)))

            for en, _ in enumerate(reference.edges):
                for i in range(1, order):
                    dofs.append(BernsteinFunctional(
                        reference, reference.sub_entity(1, en), i, order, (1, en)))

            for fn, _ in enumerate(reference.faces):
                for i in range(1, order):
                    for j in range(1, order - i):
                        dofs.append(BernsteinFunctional(
                            reference, reference.sub_entity(2, fn), index(i, j), order, (2, fn)))

            if reference.name == "tetrahedon":
                for i in range(1, order):
                    for j in range(1, order - i):
                        for k in range(1, order - i - j):
                            dofs.append(BernsteinFunctional(
                                reference, reference, index(i, j, k), order, (3, 0)))

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)

    names = ["Bernstein", "Bernstein-Bezier"]
    references = ["interval", "triangle", "tetrahedron"]
    min_order = 0
    continuity = "C0"



"""Piecewise basis function classes."""

from __future__ import annotations

import typing

import sympy

from .functions import (AnyFunction, FunctionInput, SympyFormat, ValuesToSubstitute, VectorFunction,
                        _to_sympy_format, parse_function_input)
from .geometry import (PointType, SetOfPoints, SetOfPointsInput, parse_set_of_points_input,
                       point_in_quadrilateral, point_in_tetrahedron, point_in_triangle)
from .references import Reference
from .symbols import AxisVariables, AxisVariablesNotSingle, t, x


class PiecewiseFunction(AnyFunction):
    """A piecewise function."""

    _pieces: typing.Dict[SetOfPoints, AnyFunction]

    def __init__(
        self, pieces: typing.Dict[SetOfPointsInput, FunctionInput], tdim: int
    ):
        """Create a piecewise function.

        Args:
            pieces: The pieces of the function
            tdim: The topological dimension
        """
        self._pieces = {parse_set_of_points_input(shape): parse_function_input(f)
                        for shape, f in pieces.items()}
        self.tdim = tdim

        assert len(self._pieces) > 0
        self.first_piece = list(self._pieces.values())[0]
        if self.first_piece.is_scalar:
            for f in self._pieces.values():
                assert f.is_scalar
            super().__init__(scalar=True)
        elif self.first_piece.is_vector:
            for f in self._pieces.values():
                assert f.is_vector
            super().__init__(vector=True)
        else:
            for f in self._pieces.values():
                assert f.is_matrix
            super().__init__(matrix=True)

    def __len__(self):
        """Get the length of the vector."""
        if not self.is_vector:
            raise TypeError(f"object of type '{self.__class__.__name__}' has no len()")
        return len(self.first_piece)

    def as_sympy(self) -> SympyFormat:
        """Convert to a sympy expression.

        Returns:
            A Sympy expression
        """
        for f in self._pieces.values():
            if f != self.first_piece:
                break
        else:
            return self.first_piece.as_sympy()
        out = {}
        for shape, f in self._pieces.items():
            fs = f.as_sympy()
            assert isinstance(fs, (sympy.core.expr.Expr, tuple,
                                   sympy.matrices.dense.MutableDenseMatrix))
            out[tuple(shape)] = fs
        return out

    def as_tex(self) -> str:
        """Convert to a TeX expression.

        Returns:
            A TeX string
        """
        out = "\\begin{cases}\n"
        joiner = ""
        for shape, f in self._pieces.items():
            out += joiner
            joiner = "\\\\"
            out += f.as_tex().replace("\\frac", "\\tfrac")
            out += "&\\text{in }\\operatorname{"
            if self.tdim == 2:
                if len(shape) == 3:
                    out += "Triangle"
                elif len(shape) == 4:
                    out += "Quadrilateral"
                else:
                    raise ValueError("Unsupported shape")
            elif self.tdim == 3:
                if len(shape) == 4:
                    out += "Tetrahedron"
                else:
                    raise ValueError("Unsupported shape")
            else:
                raise ValueError("Unsupported shape")
            out += "}"
            out += f"({shape})"
        out += "\\end{cases}"
        return out

    def get_piece(self, point: PointType) -> AnyFunction:
        """Get a piece of the function.

        Args:
            point: The point to get the piece at

        Returns:
            The piece of the function that is valid at that point
        """
        if self.tdim == 2:
            for cell, value in self._pieces.items():
                if len(cell) == 3:
                    if point_in_triangle(point[:2], cell):
                        return value
                elif len(cell) == 4:
                    if point_in_quadrilateral(point[:2], cell):
                        return value
                else:
                    raise ValueError("Unsupported cell")
        if self.tdim == 3:
            for cell, value in self._pieces.items():
                if len(cell) == 4:
                    if point_in_tetrahedron(point, cell):
                        return value
                else:
                    raise ValueError("Unsupported cell")

        raise NotImplementedError("Evaluation of piecewise functions outside domain not supported.")

    def __getitem__(self, key) -> PiecewiseFunction:
        """Get a component or slice of the function."""
        return PiecewiseFunction(
            {shape: f.__getitem__(key) for shape, f in self._pieces.items()}, self.tdim)

    def __eq__(self, other: typing.Any) -> bool:
        """Check if two functions are equal."""
        if not isinstance(other, PiecewiseFunction):
            return False

        if self.tdim != other.tdim:
            return False
        for (shape0, f0), (shape1, f1) in zip(self._pieces.items(), other._pieces.items()):
            assert shape0 == shape1
            if f0 != f1:
                return False
        return True

    @property
    def pieces(self) -> typing.Dict[SetOfPoints, AnyFunction]:
        """Get the pieces of the function.

        Returns:
            The function pieces
        """
        return self._pieces

    def __add__(self, other: typing.Any) -> PiecewiseFunction:
        """Add."""
        if isinstance(other, PiecewiseFunction):
            new_pieces: typing.Dict[SetOfPointsInput, FunctionInput] = {}
            for (shape0, f0), (shape1, f1) in zip(self._pieces.items(), other._pieces.items()):
                assert shape0 == shape1
                new_pieces[shape0] = f0 + f1
            return PiecewiseFunction(new_pieces, self.tdim)
        return PiecewiseFunction(
            {shape: f + other for shape, f in self._pieces.items()}, self.tdim)

    def __radd__(self, other: typing.Any) -> PiecewiseFunction:
        """Add."""
        if isinstance(other, PiecewiseFunction):
            new_pieces: typing.Dict[SetOfPointsInput, FunctionInput] = {}
            for (shape0, f0), (shape1, f1) in zip(self._pieces.items(), other._pieces.items()):
                assert shape0 == shape1
                new_pieces[shape0] = f1 + f0
            return PiecewiseFunction(new_pieces, self.tdim)
        return PiecewiseFunction(
            {shape: other + f for shape, f in self._pieces.items()}, self.tdim)

    def __sub__(self, other: typing.Any) -> PiecewiseFunction:
        """Subtract."""
        if isinstance(other, PiecewiseFunction):
            new_pieces: typing.Dict[SetOfPointsInput, FunctionInput] = {}
            for (shape0, f0), (shape1, f1) in zip(self._pieces.items(), other._pieces.items()):
                assert shape0 == shape1
                new_pieces[shape0] = f0 - f1
            return PiecewiseFunction(new_pieces, self.tdim)
        return PiecewiseFunction(
            {shape: f - other for shape, f in self._pieces.items()}, self.tdim)

    def __rsub__(self, other: typing.Any) -> PiecewiseFunction:
        """Subtract."""
        if isinstance(other, PiecewiseFunction):
            new_pieces: typing.Dict[SetOfPointsInput, FunctionInput] = {}
            for (shape0, f0), (shape1, f1) in zip(self._pieces.items(), other._pieces.items()):
                assert shape0 == shape1
                new_pieces[shape0] = f1 - f0
            return PiecewiseFunction(new_pieces, self.tdim)
        return PiecewiseFunction(
            {shape: other - f for shape, f in self._pieces.items()}, self.tdim)

    def __truediv__(self, other: typing.Any) -> PiecewiseFunction:
        """Divide."""
        if isinstance(other, PiecewiseFunction):
            new_pieces: typing.Dict[SetOfPointsInput, FunctionInput] = {}
            for (shape0, f0), (shape1, f1) in zip(self._pieces.items(), other._pieces.items()):
                assert shape0 == shape1
                new_pieces[shape0] = f0 / f1
            return PiecewiseFunction(new_pieces, self.tdim)
        return PiecewiseFunction(
            {shape: f / other for shape, f in self._pieces.items()}, self.tdim)

    def __rtruediv__(self, other: typing.Any) -> PiecewiseFunction:
        """Divide."""
        if isinstance(other, PiecewiseFunction):
            new_pieces: typing.Dict[SetOfPointsInput, FunctionInput] = {}
            for (shape0, f0), (shape1, f1) in zip(self._pieces.items(), other._pieces.items()):
                assert shape0 == shape1
                new_pieces[shape0] = f1 / f0
            return PiecewiseFunction(new_pieces, self.tdim)
        return PiecewiseFunction(
            {shape: other / f for shape, f in self._pieces.items()}, self.tdim)

    def __mul__(self, other: typing.Any) -> PiecewiseFunction:
        """Multiply."""
        if isinstance(other, PiecewiseFunction):
            new_pieces: typing.Dict[SetOfPointsInput, FunctionInput] = {}
            for (shape0, f0), (shape1, f1) in zip(self._pieces.items(), other._pieces.items()):
                assert shape0 == shape1
                new_pieces[shape0] = f0 * f1
            return PiecewiseFunction(new_pieces, self.tdim)
        return PiecewiseFunction(
            {shape: f * other for shape, f in self._pieces.items()}, self.tdim)

    def __rmul__(self, other: typing.Any) -> PiecewiseFunction:
        """Multiply."""
        if isinstance(other, PiecewiseFunction):
            new_pieces: typing.Dict[SetOfPointsInput, FunctionInput] = {}
            for (shape0, f0), (shape1, f1) in zip(self._pieces.items(), other._pieces.items()):
                assert shape0 == shape1
                new_pieces[shape0] = f1 * f0
            return PiecewiseFunction(new_pieces, self.tdim)
        return PiecewiseFunction(
            {shape: other * f for shape, f in self._pieces.items()}, self.tdim)

    def __matmul__(self, other: typing.Any) -> PiecewiseFunction:
        """Multiply."""
        if isinstance(other, PiecewiseFunction):
            new_pieces: typing.Dict[SetOfPointsInput, FunctionInput] = {}
            for (shape0, f0), (shape1, f1) in zip(self._pieces.items(), other._pieces.items()):
                assert shape0 == shape1
                new_pieces[shape0] = f0 @ f1
            return PiecewiseFunction(new_pieces, self.tdim)
        return PiecewiseFunction(
            {shape: f @ other for shape, f in self._pieces.items()}, self.tdim)

    def __rmatmul__(self, other: typing.Any) -> PiecewiseFunction:
        """Multiply."""
        if isinstance(other, PiecewiseFunction):
            new_pieces: typing.Dict[SetOfPointsInput, FunctionInput] = {}
            for (shape0, f0), (shape1, f1) in zip(self._pieces.items(), other._pieces.items()):
                assert shape0 == shape1
                new_pieces[shape0] = f1 @ f0
            return PiecewiseFunction(new_pieces, self.tdim)
        return PiecewiseFunction(
            {shape: other @ f for shape, f in self._pieces.items()}, self.tdim)

    def __pow__(self, other: typing.Any) -> PiecewiseFunction:
        """Raise to a power."""
        if isinstance(other, PiecewiseFunction):
            new_pieces: typing.Dict[SetOfPointsInput, FunctionInput] = {}
            for (shape0, f0), (shape1, f1) in zip(self._pieces.items(), other._pieces.items()):
                assert shape0 == shape1
                new_pieces[shape0] = f0 ** f1
            return PiecewiseFunction(new_pieces, self.tdim)
        return PiecewiseFunction(
            {shape: f ** other for shape, f in self._pieces.items()}, self.tdim)

    def __neg__(self) -> PiecewiseFunction:
        """Negate."""
        return PiecewiseFunction({shape: -f for shape, f in self._pieces.items()}, self.tdim)

    def subs(
        self, vars: AxisVariables, values: ValuesToSubstitute
    ) -> PiecewiseFunction:
        """Substitute values into the function.

        Args:
            vars: The variables to substitute out
            values: The value to substitute in

        Returns:
            The substituted function
        """
        if isinstance(values, AnyFunction):
            values = values.as_sympy()

        if isinstance(values, (tuple, list)) and len(values) == self.tdim:
            for i in values:
                i_s = _to_sympy_format(i)
                assert isinstance(i_s, sympy.core.expr.Expr)
                if not i_s.is_constant():
                    break
            else:
                return self.get_piece(tuple(values)).subs(vars, values)

        return PiecewiseFunction(
            {shape: f.subs(vars, values) for shape, f in self._pieces.items()}, self.tdim)

    def diff(self, variable: sympy.core.symbol.Symbol) -> PiecewiseFunction:
        """Differentiate the function.

        Args:
            variable: The variable to differentiate with respect to

        Returns:
            The differentiated function
        """
        return PiecewiseFunction(
            {shape: f.diff(variable) for shape, f in self._pieces.items()}, self.tdim)

    def directional_derivative(self, direction: PointType) -> PiecewiseFunction:
        """Compute a directional derivative.

        Args:
            direction: The diection

        Returns:
            The directional derivative
        """
        return PiecewiseFunction(
            {shape: f.directional_derivative(direction)
             for shape, f in self._pieces.items()}, self.tdim)

    def jacobian_component(self, component: typing.Tuple[int, int]) -> PiecewiseFunction:
        """Compute a component of the jacobian.

        Args:
            component: The component

        Returns:
            The component of the jacobian
        """
        return PiecewiseFunction(
            {shape: f.jacobian_component(component) for shape, f in self._pieces.items()},
            self.tdim)

    def jacobian(self, dim: int) -> PiecewiseFunction:
        """Compute the jacobian.

        Args:
            dim: The topological dimension of the cell

        Returns:
            The jacobian
        """
        return PiecewiseFunction(
            {shape: f.jacobian(dim) for shape, f in self._pieces.items()}, self.tdim)

    def dot(self, other_in: FunctionInput) -> PiecewiseFunction:
        """Compute the dot product with another function.

        Args:
            other_in: The function to multiply with

        Returns:
            The product
        """
        return PiecewiseFunction(
            {shape: f.dot(other_in) for shape, f in self._pieces.items()}, self.tdim)

    def cross(self, other_in: FunctionInput) -> PiecewiseFunction:
        """Compute the cross product with another function.

        Args:
            other_in: The function to multiply with

        Returns:
            The cross product
        """
        return PiecewiseFunction(
            {shape: f.cross(other_in) for shape, f in self._pieces.items()}, self.tdim)

    def div(self) -> PiecewiseFunction:
        """Compute the div of the function.

        Returns:
            The divergence
        """
        return PiecewiseFunction(
            {shape: f.div() for shape, f in self._pieces.items()}, self.tdim)

    def grad(self, dim: int) -> PiecewiseFunction:
        """Compute the grad of the function.

        Returns:
            The gradient
        """
        return PiecewiseFunction(
            {shape: f.grad(dim) for shape, f in self._pieces.items()}, self.tdim)

    def curl(self) -> PiecewiseFunction:
        """Compute the curl of the function.

        Returns:
            The curl
        """
        return PiecewiseFunction(
            {shape: f.curl() for shape, f in self._pieces.items()}, self.tdim)

    def norm(self) -> PiecewiseFunction:
        """Compute the norm of the function.

        Returns:
            The norm
        """
        return PiecewiseFunction(
            {shape: f.norm() for shape, f in self._pieces.items()}, self.tdim)

    def integral(self, domain: Reference, vars: AxisVariablesNotSingle = t) -> AnyFunction:
        """Compute the integral of the function.

        Args:
            domain: The domain of the integral
            vars: The variables to integrate with respect to

        Returns:
            The integral
        """
        # TODO: Add check that the domain is a subset of one piece
        # TODO: Add integral over multiple pieces
        p = self.get_piece(domain.midpoint())
        return p.integral(domain, vars)

    def det(self) -> PiecewiseFunction:
        """Compute the determinant.

        Returns:
            The deteminant
        """
        if not self.is_matrix:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute 'det'")
        return PiecewiseFunction(
            {shape: f.det() for shape, f in self._pieces.items()}, self.tdim)

    def transpose(self) -> PiecewiseFunction:
        """Compute the transpose.

        Returns:
            The transpose
        """
        if not self.is_matrix:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute 'transpose'")
        return PiecewiseFunction(
            {shape: f.transpose() for shape, f in self._pieces.items()}, self.tdim)

    def map_pieces(self, fwd_map: PointType):
        """Map the function's pieces.

        Args:
            fwd_map: The map from the reference cell to a physical cell

        Returns:
            The mapped pieces
        """
        new_pieces: typing.Dict[SetOfPointsInput, FunctionInput] = {}
        for shape, f in self._pieces.items():
            nshape = []
            for v in shape:
                pt = VectorFunction(fwd_map).subs(x, v).as_sympy()
                assert isinstance(pt, tuple)
                nshape.append(pt)
            new_pieces[tuple(nshape)] = f
        self._pieces = {parse_set_of_points_input(shape): parse_function_input(f)
                        for shape, f in new_pieces.items()}

    @property
    def shape(self) -> typing.Tuple[int, ...]:
        """Get the value shape of the function.

        Returns:
            The value shape
        """
        return self.first_piece.shape

    def plot_values(
        self, reference: Reference, img: typing.Any,
        value_scale: sympy.core.expr.Expr = sympy.Integer(1), n: int = 6
    ):
        """Plot the function's values.

        Args:
            reference: The reference cell
            img: The image to plot on
            value_scale: The scale factor for the function values
            n: The number of points per side for plotting
        """
        from .create import create_reference
        from .plotting import Picture
        assert isinstance(img, Picture)

        for shape, f in self._pieces.items():
            if self.tdim == 2:
                if len(shape) == 3:
                    ref = create_reference("triangle", shape)
                elif len(shape) == 4:
                    ref = create_reference("quadrilateral", shape)
                else:
                    raise ValueError("Unsupported cell type")
            elif self.tdim == 3:
                if len(shape) == 4:
                    ref = create_reference("tetrahedron", shape)
                else:
                    raise ValueError("Unsupported cell type")
            else:
                raise ValueError("Unsupported tdim")
            f.plot_values(ref, img, value_scale, n // 2)

    def with_floats(self) -> AnyFunction:
        """Return a version the function with floats as coefficients.

        Returns:
            A version the function with floats as coefficients
        """
        return PiecewiseFunction(
            {shape: f.with_floats() for shape, f in self._pieces.items()}, self.tdim)



"""Morley-Wang-Xu elements on simplices.

This element's definition appears in https://doi.org/10.1090/S0025-5718-2012-02611-1
(Wang, Xu, 2013)
"""

import typing

from ..finite_element import CiarletElement
from ..functionals import (IntegralAgainst, IntegralOfDirectionalMultiderivative, ListOfFunctionals,
                           PointEvaluation)
from ..functions import FunctionInput
from ..polynomials import polynomial_set_1d
from ..references import Reference


class MorleyWangXu(CiarletElement):
    """Morley-Wang-Xu finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        assert order <= reference.tdim
        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_1d(reference.tdim, order)

        dofs: ListOfFunctionals = []

        if order == 1:
            if reference.tdim == 1:
                for v_n, v in enumerate(reference.vertices):
                    dofs.append(PointEvaluation(reference, v, entity=(0, v_n)))
            else:
                dim = reference.tdim - 1
                for facet_n in range(reference.sub_entity_count(dim)):
                    facet = reference.sub_entity(dim, facet_n)
                    dofs.append(IntegralAgainst(reference, facet, 1 / facet.jacobian(),
                                                entity=(dim, facet_n)))
        elif order == 2:
            if reference.tdim == 2:
                for v_n, v in enumerate(reference.vertices):
                    dofs.append(PointEvaluation(reference, v, entity=(0, v_n)))
            else:
                dim = reference.tdim - 2
                for ridge_n in range(reference.sub_entity_count(dim)):
                    ridge = reference.sub_entity(dim, ridge_n)
                    dofs.append(IntegralAgainst(reference, ridge, 1 / ridge.jacobian(),
                                                entity=(dim, ridge_n)))
            dim = reference.tdim - 1
            for facet_n in range(reference.sub_entity_count(dim)):
                facet = reference.sub_entity(dim, facet_n)
                dofs.append(IntegralOfDirectionalMultiderivative(
                    reference, facet, (facet.normal(), ), (1, ), (dim, facet_n),
                    scale=1 / facet.jacobian()))
        else:
            assert order == reference.tdim == 3
            for v_n, v in enumerate(reference.vertices):
                dofs.append(PointEvaluation(reference, v, entity=(0, v_n)))
            for e_n, vs in enumerate(reference.sub_entities(1)):
                subentity = reference.sub_entity(1, e_n)
                volume = subentity.jacobian()
                normals = []
                for f_n, f_vs in enumerate(reference.sub_entities(2)):
                    if vs[0] in f_vs and vs[1] in f_vs:
                        face = reference.sub_entity(2, f_n)
                        normals.append(face.normal())
                for orders in [(1, 0), (0, 1)]:
                    dofs.append(IntegralOfDirectionalMultiderivative(
                        reference, subentity, tuple(normals), orders, (1, e_n),
                        scale=1 / volume))
            for f_n, vs in enumerate(reference.sub_entities(2)):
                subentity = reference.sub_entity(2, f_n)
                volume = subentity.jacobian()
                dofs.append(IntegralOfDirectionalMultiderivative(
                    reference, subentity, (subentity.normal(), ), (2, ), (2, f_n),
                    scale=1 / volume))

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)

    names = ["Morley-Wang-Xu", "MWX"]
    references = ["interval", "triangle", "tetrahedron"]
    min_order = 1
    max_order = {"interval": 1, "triangle": 2, "tetrahedron": 3}
    continuity = "C{order}"



"""Bell elements on triangle.

This element's definition is given in https://doi.org/10.1002/nme.1620010108 (Bell, 1969)
"""

import typing

from ..finite_element import CiarletElement
from ..functionals import (DerivativePointEvaluation, ListOfFunctionals,
                           NormalDerivativeIntegralMoment, PointEvaluation)
from ..functions import FunctionInput
from ..moments import make_integral_moment_dofs
from ..polynomials import polynomial_set_1d
from ..references import Reference
from .lagrange import Lagrange


class Bell(CiarletElement):
    """Bell finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        assert reference.name == "triangle"
        assert order == 5
        dofs: ListOfFunctionals = []
        for v_n, v in enumerate(reference.vertices):
            dofs.append(PointEvaluation(reference, v, entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(reference, v, (1, 0), entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(reference, v, (0, 1), entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(reference, v, (2, 0), entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(reference, v, (1, 1), entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(reference, v, (0, 2), entity=(0, v_n)))
        dofs += make_integral_moment_dofs(
            reference,
            edges=(NormalDerivativeIntegralMoment, Lagrange, 0, {"variant": variant}),
        )
        self.variant = variant

        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_1d(reference.tdim, order)

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["Bell"]
    references = ["triangle"]
    min_order = 5
    max_order = 5
    continuity = "C1"



"""Nedelec elements on prisms."""

import typing

from ..finite_element import CiarletElement
from ..functionals import IntegralAgainst, IntegralMoment, ListOfFunctionals, TangentIntegralMoment
from ..functions import FunctionInput
from ..moments import make_integral_moment_dofs
from ..polynomials import Hcurl_polynomials, polynomial_set_1d, polynomial_set_vector
from ..references import Reference
from ..symbols import x
from .lagrange import Lagrange, VectorLagrange
from .q import RaviartThomas as QRT


class Nedelec(CiarletElement):
    """Nedelec Hcurl finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        from .. import create_reference

        poly: typing.List[FunctionInput] = []
        poly += [
            (i[0] * j, i[1] * j, 0)
            for i in polynomial_set_vector(2, 2, order - 1) + Hcurl_polynomials(2, 2, order)
            for j in polynomial_set_1d(1, order, x[2:])]
        poly += [(0, 0, i * j)
                 for i in polynomial_set_1d(2, order, x[:2])
                 for j in polynomial_set_1d(1, order - 1, x[2:])]

        dofs: ListOfFunctionals = make_integral_moment_dofs(
            reference,
            edges=(TangentIntegralMoment, Lagrange, order - 1,
                   {"variant": variant}),
            faces={"triangle": (IntegralMoment, VectorLagrange, order - 2,
                                "covariant", {"variant": variant}),
                   "quadrilateral": (IntegralMoment, QRT, order - 1, "covariant",
                                     {"variant": variant})},
        )

        triangle = create_reference("triangle")
        interval = create_reference("interval")

        if order >= 2:
            space1 = VectorLagrange(triangle, order - 2, variant)
            space2 = Lagrange(interval, order - 2, variant)

            if order > 2:
                raise NotImplementedError()
            # TODO: correct these for order > 2
            for i in range(space1.space_dim):
                for j in range(space2.space_dim):
                    f = (space2.get_basis_function(j) * space1.get_basis_function(i)[0],
                         space2.get_basis_function(j) * space1.get_basis_function(i)[1],
                         0)
                    dofs.append(IntegralAgainst(
                        reference, reference, f, entity=(3, 0), mapping="covariant"))

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["Nedelec", "Ncurl"]
    references = ["prism"]
    min_order = 1
    max_order = 2
    continuity = "H(curl)"



"""TiNiest Tensor product (TNT) elements.

These elements' definitions appear in https://doi.org/10.1090/S0025-5718-2013-02729-9
(Cockburn, Qiu, 2013)
"""

import typing
from itertools import product

import sympy

from ..finite_element import CiarletElement
from ..functionals import (DerivativeIntegralMoment, IntegralAgainst, ListOfFunctionals,
                           NormalIntegralMoment, PointEvaluation, TangentIntegralMoment)
from ..functions import FunctionInput, ScalarFunction, VectorFunction
from ..moments import make_integral_moment_dofs
from ..polynomials import orthogonal_basis, quolynomial_set_1d, quolynomial_set_vector
from ..references import Reference
from ..symbols import t, x
from .q import Q


def p(k: int, v: sympy.core.symbol.Symbol) -> ScalarFunction:
    """Return the kth Legendre polynomial.

    Args:
        k: k
        v: The variable to use

    Returns:
        The kth Legendre polynomial
    """
    return orthogonal_basis("interval", k, 0, [v])[0][-1]


def b(k: int, v: sympy.core.symbol.Symbol) -> ScalarFunction:
    """Return the function B_k.

    This function is defined on page 4 (606) of
    https://doi.org/10.1090/S0025-5718-2013-02729-9 (Cockburn, Qiu, 2013).

    Args:
        k: k
        v: The variable to use

    Returns:
        The function B_k
    """
    if k == 1:
        return ScalarFunction(0)
    return (p(k, v) - p(k - 2, v)) / (4 + k - 2)


class TNT(CiarletElement):
    """TiNiest Tensor scalar finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        poly: typing.List[FunctionInput] = []
        poly += quolynomial_set_1d(reference.tdim, order)
        if reference.tdim == 2:
            for i in range(2):
                variables = [x[j] for j in range(2) if j != i]
                for f in [1 - variables[0], variables[0]]:
                    poly.append(f * b(order + 1, x[i]))

        elif reference.tdim == 3:
            for i in range(3):
                variables = [x[j] for j in range(3) if j != i]
                for f0 in [1 - variables[0], variables[0]]:
                    for f1 in [1 - variables[1], variables[1]]:
                        poly.append(f0 * f1 * b(order + 1, x[i]))

        dofs: ListOfFunctionals = []
        for i, v in enumerate(reference.vertices):
            dofs.append(PointEvaluation(reference, v, entity=(0, i)))

        for i in range(1, order + 1):
            f = i * t[0] ** (i - 1)
            for edge_n in range(reference.sub_entity_count(1)):
                edge = reference.sub_entity(1, edge_n)
                dofs.append(IntegralAgainst(
                    reference, edge, f, entity=(1, edge_n), mapping="identity"))

        for i in range(1, order):
            for j in range(1, order):
                f = t[0] ** i * (t[0] - 1) * t[1] ** j * (t[1] - 1)
                delta_f = (f.diff(t[0]).diff(t[0]) + f.diff(t[1]).diff(t[1])).expand()
                for face_n in range(reference.sub_entity_count(2)):
                    face = reference.sub_entity(2, face_n)
                    dofs.append(IntegralAgainst(
                        reference, face, delta_f, entity=(2, face_n), mapping="identity"))

        if reference.tdim == 3:
            dummy_dof = PointEvaluation(reference, reference.midpoint(), (3, 0))
            for ii in product(range(1, order), repeat=3):
                f = sympy.Integer(1)
                for j, k in zip(ii, x):
                    f *= k ** j * (k - 1)
                grad_f = tuple(sympy.S(j).expand() for j in ScalarFunction(f).grad(3))
                dofs.append(DerivativeIntegralMoment(
                    reference, reference, 1, grad_f, dummy_dof, entity=(3, 0), mapping="identity"))

        super().__init__(
            reference, order, poly, dofs, reference.tdim, 1
        )
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["tiniest tensor", "TNT"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 1
    continuity = "C0"


class TNTcurl(CiarletElement):
    """TiNiest Tensor Hcurl finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        poly: typing.List[FunctionInput] = []
        poly += quolynomial_set_vector(reference.tdim, reference.tdim, order)
        if reference.tdim == 2:
            for ii in product([0, 1], repeat=2):
                if sum(ii) != 0:
                    poly.append(tuple(sympy.S(j).expand() for j in [
                        p(order, ii[0] * x[0]) * b(order + 1, ii[1] * x[1]),
                        -b(order + 1, ii[0] * x[0]) * p(order, ii[1] * x[1])]))
        else:
            face_poly = []
            for ii in product([0, 1], repeat=2):
                if sum(ii) != 0:
                    face_poly.append(tuple(sympy.S(j).expand() for j in [
                        b(order + 1, ii[0] * t[0]) * p(order, ii[1] * t[1]),
                        p(order, ii[0] * t[0]) * b(order + 1, ii[1] * t[1])]))
            for lamb_n in [(x[0], 0, 0), (1 - x[0], 0, 0),
                           (0, x[1], 0), (0, 1 - x[1], 0),
                           (0, 0, x[2]), (0, 0, 1 - x[2])]:
                variables = tuple(i for i, j in enumerate(lamb_n) if j == 0)
                for pf in face_poly:
                    psub = VectorFunction(pf).subs(t[:2], [x[j] for j in variables])
                    pc = VectorFunction(lamb_n).cross(VectorFunction([
                        psub[variables.index(i)] if i in variables else 0 for i in range(3)]))
                    poly.append(pc)

        dofs: ListOfFunctionals = []
        dofs += make_integral_moment_dofs(
            reference,
            edges=(TangentIntegralMoment, Q, order, {"variant": variant}),
        )

        # Face moments
        face_moments = []
        for ii in product(range(order + 1), repeat=2):
            if sum(ii) > 0:
                f = x[0] ** ii[0] * x[1] ** ii[1]
                grad_f = ScalarFunction(f).grad(2)
                grad_f2 = VectorFunction((grad_f[1], -grad_f[0])).subs(x[:2], tuple(t[:2]))
                face_moments.append(grad_f2)

        for i in range(2, order + 1):
            for j in range(2, order + 1):
                face_moments.append(VectorFunction((
                    t[1] ** (j - 1) * (1 - t[1]) * t[0] ** (i - 2) * (i * t[0] - i + 1),
                    -t[0] ** (i - 1) * (1 - t[0]) * t[1] ** (j - 2) * (j - 1 - j * t[1]))))
        if reference.tdim == 2:
            for f in face_moments:
                dofs.append(IntegralAgainst(
                    reference, reference, f, entity=(2, 0), mapping="contravariant"))
        elif reference.tdim == 3:
            for face_n in range(6):
                face = reference.sub_entity(2, face_n)
                for f in face_moments:
                    dofs.append(IntegralAgainst(
                        reference, face, f, entity=(2, face_n), mapping="contravariant"))

        # Interior Moments
        if reference.tdim == 3:
            for i in range(1, order):
                for j in range(1, order):
                    for k in range(order + 1):
                        f = (x[0] ** k * x[1] ** i * (1 - x[1]) * x[2] ** j * (1 - x[2]), 0, 0)
                        dofs.append(IntegralAgainst(
                            reference, reference, VectorFunction(f).curl().curl(), entity=(3, 0),
                            mapping="covariant"))

                        f = (0, x[1] ** k * x[0] ** i * (1 - x[0]) * x[2] ** j * (1 - x[2]), 0)
                        dofs.append(IntegralAgainst(
                            reference, reference, VectorFunction(f).curl().curl(), entity=(3, 0),
                            mapping="covariant"))

                        if k in [0, 2]:
                            f = (0, 0,  x[2] ** k * x[0] ** i * (1 - x[0]) * x[1] ** j * (1 - x[1]))
                            dofs.append(IntegralAgainst(
                                reference, reference, VectorFunction(f).curl().curl(),
                                entity=(3, 0), mapping="covariant"))

            for i in range(2, order + 1):
                for j in range(2, order + 1):
                    for k in range(2, order + 1):
                        f = x[0] ** (i - 1) * x[0] ** i
                        f *= x[1] ** (j - 1) * x[1] ** j
                        f *= x[2] ** (k - 1) * x[2] ** k
                        grad_f = ScalarFunction(f).grad(3)
                        dofs.append(IntegralAgainst(
                            reference, reference, grad_f, entity=(3, 0), mapping="contravariant"))

        super().__init__(
            reference, order, poly, dofs, reference.tdim, reference.tdim
        )
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["tiniest tensor Hcurl", "TNTcurl"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 1
    continuity = "H(curl)"


class TNTdiv(CiarletElement):
    """TiNiest Tensor Hdiv finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        poly: typing.List[FunctionInput] = []
        poly += quolynomial_set_vector(reference.tdim, reference.tdim, order)
        if reference.tdim == 2:
            for ii in product([0, 1], repeat=2):
                if sum(ii) != 0:
                    poly.append(tuple(sympy.S(j).expand() for j in [
                        b(order + 1, ii[0] * x[0]) * p(order, ii[1] * x[1]),
                        p(order, ii[0] * x[0]) * b(order + 1, ii[1] * x[1]),
                    ]))
        else:
            for ii in product([0, 1], repeat=3):
                if sum(ii) != 0:
                    poly.append((
                        b(order + 1,
                          ii[0] * x[0]) * p(order, ii[1] * x[1]) * p(order, ii[2] * x[2]),
                        p(order,
                          ii[0] * x[0]) * b(order + 1, ii[1] * x[1]) * p(order, ii[2] * x[2]),
                        p(order,
                          ii[0] * x[0]) * p(order, ii[1] * x[1]) * b(order + 1, ii[2] * x[2]),
                    ))

        dofs: ListOfFunctionals = []
        dofs += make_integral_moment_dofs(
            reference,
            facets=(NormalIntegralMoment, Q, order, {"variant": variant}),
        )

        for ii in product(range(order + 1), repeat=reference.tdim):
            if sum(ii) > 0:
                if reference.tdim == 2:
                    f = x[0] ** ii[0] * x[1] ** ii[1]
                else:
                    f = x[0] ** ii[0] * x[1] ** ii[1] * x[2] ** ii[2]
                grad_f = ScalarFunction(f).grad(reference.tdim)
                dofs.append(IntegralAgainst(
                    reference, reference, grad_f, entity=(reference.tdim, 0), mapping="covariant"))

        if reference.tdim == 2:
            for i in range(2, order + 1):
                for j in range(2, order + 1):
                    f = (x[0] ** (i - 1) * (1 - x[0]) * x[1] ** (j - 2) * (j - 1 - j * x[1]),
                         x[1] ** (j - 1) * (1 - x[1]) * x[0] ** (i - 2) * (i * x[0] - i + 1))
                    dofs.append(IntegralAgainst(
                        reference, reference, f, entity=(reference.tdim, 0), mapping="covariant"))
        if reference.tdim == 3:
            for i in range(2, order + 1):
                for j in range(2, order + 1):
                    for k in range(order + 1):
                        f = (
                            x[2] ** k * x[0] ** (i - 1) * (1 - x[0]) * x[2] ** (j - 2) * (
                                j - 1 - j * x[1]),
                            x[2] ** k * x[1] ** (j - 1) * (1 - x[1]) * x[0] ** (i - 2) * (
                                i * x[0] - i + 1),
                            0
                        )
                        dofs.append(IntegralAgainst(
                            reference, reference, f, entity=(reference.tdim, 0),
                            mapping="covariant"))
                        f = (
                            x[1] ** k * x[0] ** (i - 1) * (1 - x[0]) * x[2] ** (j - 2) * (
                                j - 1 - j * x[2]),
                            0,
                            x[1] ** k * x[2] ** (j - 1) * (1 - x[2]) * x[0] ** (i - 2) * (
                                i * x[0] - i + 1)
                        )
                        dofs.append(IntegralAgainst(
                            reference, reference, f, entity=(reference.tdim, 0),
                            mapping="covariant"))
                        if k in [0, 2]:
                            f = (
                                0,
                                x[0] ** k * x[1] ** (i - 1) * (1 - x[1]) * x[2] ** (j - 2) * (
                                    j - 1 - j * x[2]),
                                x[0] ** k * x[2] ** (j - 1) * (1 - x[2]) * x[1] ** (i - 2) * (
                                    i * x[1] - i + 1)
                            )
                            dofs.append(IntegralAgainst(
                                reference, reference, f, entity=(reference.tdim, 0),
                                mapping="covariant"))

        super().__init__(
            reference, order, poly, dofs, reference.tdim, reference.tdim
        )
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["tiniest tensor Hdiv", "TNTdiv"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 1
    continuity = "H(div)"



"""Rannacher-Turek elements on tensor product cells.

This element's definition appears in https://doi.org/10.1002/num.1690080202
(Rannacher, Turek, 1992)
"""

import typing

from ..finite_element import CiarletElement
from ..functionals import ListOfFunctionals, PointEvaluation
from ..functions import FunctionInput
from ..references import Reference
from ..symbols import x


class RannacherTurek(CiarletElement):
    """Rannacher-Turek finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        assert order == 1

        dofs: ListOfFunctionals = []
        for e_n, vs in enumerate(reference.sub_entities(reference.tdim - 1)):
            pt = reference.sub_entity(reference.tdim - 1, e_n).midpoint()
            dofs.append(PointEvaluation(reference, pt, entity=(reference.tdim - 1, e_n)))

        poly: typing.List[FunctionInput] = []
        if reference.name == "quadrilateral":
            poly += [1, x[0], x[1], x[0]**2 - x[1]**2]
        else:
            poly += [1, x[0], x[1], x[2], x[0]**2 - x[1]**2, x[1]**2 - x[2]**2]

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {}

    names = ["Rannacher-Turek"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 1
    max_order = 1
    continuity = "L2"



"""Test Guzman-Neilan elements."""

from itertools import combinations

import pytest
import sympy

import symfem
from symfem.elements.guzman_neilan import make_piecewise_lagrange
from symfem.symbols import x


@pytest.mark.parametrize("order", [1])
def test_guzman_neilan_triangle(order):
    e = symfem.create_element("triangle", "Guzman-Neilan", order)

    for p in e._basis[-3:]:
        for piece in p.pieces.values():
            float(piece.div().as_sympy().expand())


@pytest.mark.parametrize("order", [1, 2])
def test_guzman_neilan_tetrahedron(order):
    e = symfem.create_element("tetrahedron", "Guzman-Neilan", order)

    mid = tuple(sympy.Rational(sum(i), len(i)) for i in zip(*e.reference.vertices))
    for p in e._basis[-4:]:
        for piece in p.pieces.values():
            float(piece.div().as_sympy().expand())

        assert p.subs(x, mid) == (0, 0, 0)


@pytest.mark.parametrize("order", [1])
def test_basis_continuity_triangle(order):
    N = 5
    e = symfem.create_element("triangle", "Guzman-Neilan", order)
    third = sympy.Rational(1, 3)
    one = sympy.Integer(1)
    for pt in [(0, 0), (1, 0), (0, 1), (third, third)]:
        for f in e.get_polynomial_basis():
            value = None
            for p in f.pieces.items():
                if pt in p[0]:
                    if value is None:
                        value = p[1].subs(x, pt)
                    assert value == p[1].subs(x, pt)
    for pts in combinations([(0, 0), (1, 0), (1, 0), (third, third)], 2):
        for i in range(N + 1):
            pt = tuple(a + (b - a) * i * one / N for a, b in zip(*pts))
            for f in e.get_polynomial_basis():
                value = None
                for p in f.pieces.items():
                    if pts[0] in p[0] and pts[1] in p[0]:
                        if value is None:
                            value = p[1].subs(x, pt)
                        assert value == p[1].subs(x, pt)


@pytest.mark.parametrize("order", [1, 2])
def test_basis_continuity_tetrahedron(order):
    N = 5
    e = symfem.create_element("tetrahedron", "Guzman-Neilan", order)
    quarter = sympy.Rational(1, 4)
    one = sympy.Integer(1)
    for pt in [(0, 0, 0), (1, 0, 0), (0, 1, 0), (0, 0, 1), (quarter, quarter, quarter)]:
        for f in e.get_polynomial_basis():
            value = None
            for p in f.pieces.items():
                if pt in p[0]:
                    if value is None:
                        value = p[1].subs(x, pt)
                    assert value == p[1].subs(x, pt)
    for pts in combinations([(0, 0, 0), (1, 0, 0), (0, 1, 0), (0, 0, 1),
                             (quarter, quarter, quarter)], 2):
        for i in range(N + 1):
            pt = tuple(a + (b - a) * i * one / N for a, b in zip(*pts))
            for f in e.get_polynomial_basis():
                value = None
                for p in f.pieces.items():
                    if pts[0] in p[0] and pts[1] in p[0]:
                        if value is None:
                            value = p[1].subs(x, pt)
                        assert value == p[1].subs(x, pt)
    for pts in combinations([(0, 0, 0), (1, 0, 0), (0, 1, 0), (0, 0, 1),
                             (quarter, quarter, quarter)], 3):
        for i in range(N + 1):
            for j in range(N + 1 - i):
                pt = tuple(a + (b - a) * i * one / N + (c - a) * j * one / N
                           for a, b, c in zip(*pts))
                for f in e.get_polynomial_basis():
                    value = None
                    for p in f.pieces.items():
                        if pts[0] in p[0] and pts[1] in p[0] and pts[2] in p[0]:
                            if value is None:
                                value = p[1].subs(x, pt)
                            assert value == p[1].subs(x, pt)


@pytest.mark.parametrize("order", [1, 2])
def test_piecewise_lagrange_triangle(order):
    reference = symfem.create_reference("triangle")
    mid = tuple(sympy.Rational(sum(i), len(i)) for i in zip(*reference.vertices))
    sub_tris = [
        (reference.vertices[0], reference.vertices[1], mid),
        (reference.vertices[0], reference.vertices[2], mid),
        (reference.vertices[1], reference.vertices[2], mid)]

    N = 5
    fs = make_piecewise_lagrange(sub_tris, "triangle", order, zero_on_boundary=True)
    for e_n in range(reference.sub_entity_count(1)):
        edge = reference.sub_entity(1, e_n)
        for i in range(N + 1):
            point = edge.get_point((sympy.Rational(i, N), ))
            for f in fs:
                assert f.subs(x, point) == (0, 0)

    fs = make_piecewise_lagrange(sub_tris, "triangle", order, zero_at_centre=True)
    for f in fs:
        assert f.subs(x, mid) == (0, 0)


@pytest.mark.parametrize("order", [1, 2, 3])
def test_piecewise_lagrange_tetrahedron(order):
    reference = symfem.create_reference("tetrahedron")
    mid = tuple(sympy.Rational(sum(i), len(i)) for i in zip(*reference.vertices))
    sub_tets = [
        (reference.vertices[0], reference.vertices[1], reference.vertices[2], mid),
        (reference.vertices[0], reference.vertices[1], reference.vertices[3], mid),
        (reference.vertices[0], reference.vertices[2], reference.vertices[3], mid),
        (reference.vertices[1], reference.vertices[2], reference.vertices[3], mid)]

    N = 5
    fs = make_piecewise_lagrange(sub_tets, "tetrahedron", order, zero_on_boundary=True)
    for f_n in range(reference.sub_entity_count(2)):
        face = reference.sub_entity(2, f_n)
        for i in range(N + 1):
            for j in range(N + 1 - i):
                point = face.get_point((sympy.Rational(i, N),
                                        sympy.Rational(j, N)))
                for f in fs:
                    assert f.subs(x, point) == (0, 0, 0)

    fs = make_piecewise_lagrange(sub_tets, "tetrahedron", order, zero_at_centre=True)
    for f in fs:
        assert f.subs(x, mid) == (0, 0, 0)



"""Reduced Hsieh-Clough-Tocher elements on simplices.

This element's definition appears in https://doi.org/10.2307/2006147
(Ciarlet, 1978)
"""

import typing

import sympy

from ..finite_element import CiarletElement
from ..functionals import DerivativePointEvaluation, ListOfFunctionals, PointEvaluation
from ..functions import FunctionInput
from ..piecewise_functions import PiecewiseFunction
from ..references import Reference
from ..symbols import x


class P1Hermite(CiarletElement):
    """P1Hermite finite element."""

    def __init__(self, reference: Reference, order: int, poly: typing.List[FunctionInput]):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            poly: The polynomial basis
        """
        assert order == 3
        dofs: ListOfFunctionals = []
        for v_n, vs in enumerate(reference.vertices):
            dofs.append(PointEvaluation(reference, vs, entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(reference, vs, (1, 0), entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(reference, vs, (0, 1), entity=(0, v_n)))

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"poly": self._basis}

    names: typing.List[str] = []
    references = ["triangle"]
    min_order = 3
    max_order = 3
    continuity = "C0"


class ReducedHsiehCloughTocher(CiarletElement):
    """Reduced Hsieh-Clough-Tocher finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        from symfem import create_reference
        assert order == 3
        assert reference.name == "triangle"
        dofs: ListOfFunctionals = []
        for v_n, vs in enumerate(reference.vertices):
            dofs.append(PointEvaluation(reference, vs, entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(reference, vs, (1, 0), entity=(0, v_n)))
            dofs.append(DerivativePointEvaluation(reference, vs, (0, 1), entity=(0, v_n)))

        mid = tuple(sympy.Rational(sum(i), len(i)) for i in zip(*reference.vertices))

        subs = [
            (reference.vertices[0], reference.vertices[1], mid),
            (reference.vertices[1], reference.vertices[2], mid),
            (reference.vertices[2], reference.vertices[0], mid)]

        refs = [create_reference("triangle", vs) for vs in subs]

        polys: typing.List[typing.List[FunctionInput]] = [[], [], []]
        for i in range(order + 1):
            for j in range(order + 1 - i):
                if i != 2 or j != 1:
                    polys[0].append(x[0] ** i * x[1] ** j)
        polys[1] += [1, x[0], x[0] ** 2, x[1], x[0] * x[1], x[1] ** 2,
                     x[0] * x[1] ** 2 - x[0] ** 2 * x[1],
                     x[0] ** 3 - x[1] ** 3, x[0] ** 3 + 3 * x[0] * x[1] ** 2]
        for i in range(order + 1):
            for j in range(order + 1 - i):
                if i != 1 or j != 2:
                    polys[2].append(x[0] ** i * x[1] ** j)

        bases = []
        for r, p in zip(refs, polys):
            bf = []
            for f in P1Hermite(r, 3, p).get_basis_functions():
                bf.append(f)
            bases.append(bf)

        piece_list: typing.List[typing.Tuple[FunctionInput, ...]] = []
        piece_list.append((bases[0][0], 0, bases[2][3]))
        piece_list.append((bases[0][1], 0, bases[2][4]))
        piece_list.append((bases[0][2], 0, bases[2][5]))
        piece_list.append((bases[0][3], bases[1][0], 0))
        piece_list.append((bases[0][4], bases[1][1], 0))
        piece_list.append((bases[0][5], bases[1][2], 0))
        # TODO: are these right to remove??
        # piece_list.append((bases[0][6], bases[1][6], bases[2][6]))
        # piece_list.append((bases[0][7], bases[1][7], bases[2][7]))
        # piece_list.append((bases[0][8], bases[1][8], bases[2][8]))
        piece_list.append((0, bases[1][3], bases[2][0]))
        piece_list.append((0, bases[1][4], bases[2][1]))
        piece_list.append((0, bases[1][5], bases[2][2]))

        poly: typing.List[FunctionInput] = []
        poly += [
            PiecewiseFunction({i: j for i, j in zip(subs, p)}, 2)
            for p in piece_list
        ]

        super().__init__(
            reference, order, poly, dofs, reference.tdim, 1
        )

    names = ["reduced Hsieh-Clough-Tocher", "rHCT"]
    references = ["triangle"]
    min_order = 3
    max_order = 3
    continuity = "C1"



"""Test quadrature rules."""

import pytest
import sympy

from symfem import quadrature
from symfem.functions import ScalarFunction
from symfem.utils import allequal


@pytest.mark.parametrize("order", range(1, 7))
def test_equispaced(order):
    points, weights = quadrature.equispaced(order + 1)

    x = sympy.Symbol("x")
    poly = ScalarFunction(x)

    assert allequal(
        poly.integrate((x, 0, 1)),
        sum(i * poly.subs(x, j) for i, j in zip(weights, points)))


@pytest.mark.parametrize("order", range(1, 7))
def test_lobatto(order):
    points, weights = quadrature.lobatto(order + 1)

    x = sympy.Symbol("x")
    poly = ScalarFunction(x ** (2 * order - 1))

    assert allequal(
        poly.integrate((x, 0, 1)),
        sum(i * poly.subs(x, j) for i, j in zip(weights, points)))


@pytest.mark.parametrize("order", range(1, 3))
def test_radau(order):
    points, weights = quadrature.radau(order + 1)

    x = sympy.Symbol("x")
    poly = ScalarFunction(x ** (2 * order - 1))

    assert allequal(
        poly.integrate((x, 0, 1)),
        sum(i * poly.subs(x, j) for i, j in zip(weights, points)))


@pytest.mark.parametrize("order", range(1, 4))
def test_legendre(order):
    points, weights = quadrature.legendre(order)

    x = sympy.Symbol("x")
    poly = ScalarFunction(x ** (2 * order - 1))

    assert allequal(
        poly.integrate((x, 0, 1)),
        sum(i * poly.subs(x, j) for i, j in zip(weights, points)))



"""Q elements on tensor product cells."""

import typing
from itertools import product

import sympy

from ..finite_element import CiarletElement, FiniteElement
from ..functionals import (DotPointEvaluation, IntegralMoment, ListOfFunctionals,
                           NormalIntegralMoment, PointEvaluation, TangentIntegralMoment)
from ..functions import FunctionInput
from ..moments import make_integral_moment_dofs
from ..polynomials import (Hcurl_quolynomials, Hdiv_quolynomials, quolynomial_set_1d,
                           quolynomial_set_vector)
from ..quadrature import get_quadrature
from ..references import Reference


class Q(CiarletElement):
    """A Q element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        dofs: ListOfFunctionals = []
        if order == 0:
            dofs = [PointEvaluation(
                reference, tuple(sympy.Rational(1, 2) for i in range(reference.tdim)),
                entity=(reference.tdim, 0))]
        else:
            points, _ = get_quadrature(variant, order + 1)

            for v_n, v in enumerate(reference.vertices):
                dofs.append(PointEvaluation(reference, v, entity=(0, v_n)))
            for edim in range(1, 4):
                for e_n in range(reference.sub_entity_count(edim)):
                    entity = reference.sub_entity(edim, e_n)
                    for i in product(range(1, order), repeat=edim):
                        dofs.append(
                            PointEvaluation(
                                reference, tuple(o + sum(a[j] * points[b]
                                                         for a, b in zip(entity.axes, i[::-1]))
                                                 for j, o in enumerate(entity.origin)),
                                entity=(edim, e_n)))

        poly: typing.List[FunctionInput] = []
        poly += quolynomial_set_1d(reference.tdim, order)

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)
        self.variant = variant

    def get_tensor_factorisation(
        self
    ) -> typing.List[typing.Tuple[str, typing.List[FiniteElement], typing.List[int]]]:
        """Get the representation of the element as a tensor product.

        Returns:
            The tensor factorisation
        """
        from symfem import create_element
        interval_q = create_element("interval", "Lagrange", self.order)

        if self.order == 0:
            perm = [0]
        elif self.reference.name == "quadrilateral":
            n = self.order - 1
            perm = [0, 2] + [4 + n + i for i in range(n)]
            perm += [1, 3] + [4 + 2 * n + i for i in range(n)]
            for i in range(n):
                perm += [4 + i, 4 + 3 * n + i] + [4 + i + (4 + j) * n for j in range(n)]
        elif self.reference.name == "hexahedron":
            n = self.order - 1
            perm = [0, 4] + [8 + 2 * n + i for i in range(n)]
            perm += [2, 6] + [8 + 6 * n + i for i in range(n)]
            for i in range(n):
                perm += [8 + n + i, 8 + 9 * n + i]
                perm += [8 + 12 * n + 2 * n ** 2 + i + n * j for j in range(n)]
            perm += [1, 5] + [8 + 4 * n + i for i in range(n)]
            perm += [3, 7] + [8 + 7 * n + i for i in range(n)]
            for i in range(n):
                perm += [8 + 3 * n + i, 8 + 10 * n + i]
                perm += [8 + 12 * n + 3 * n ** 2 + i + n * j for j in range(n)]
            for i in range(n):
                perm += [8 + i, 8 + 8 * n + i]
                perm += [8 + 12 * n + n ** 2 + i + n * j for j in range(n)]
                perm += [8 + 5 * n + i, 8 + 11 * n + i]
                perm += [8 + 12 * n + 4 * n ** 2 + i + n * j for j in range(n)]
                for j in range(n):
                    perm += [8 + 12 * n + i + n * j, 8 + 12 * n + 5 * n ** 2 + i + n * j]
                    perm += [8 + 12 * n + 6 * n ** 2 + i + n * j + n ** 2 * k for k in range(n)]

        return [("scalar", [interval_q for i in range(self.reference.tdim)], perm)]

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["Q", "Lagrange", "P"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 0
    continuity = "C0"


class VectorQ(CiarletElement):
    """A vector Q element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        scalar_space = Q(reference, order, variant)
        dofs: ListOfFunctionals = []
        poly: typing.List[FunctionInput] = []
        if reference.tdim == 1:
            for p in scalar_space.dofs:
                dofs.append(PointEvaluation(reference, p.dof_point(), entity=p.entity))

            poly += quolynomial_set_1d(reference.tdim, order)
        else:
            directions = [
                tuple(1 if i == j else 0 for j in range(reference.tdim))
                for i in range(reference.tdim)
            ]
            for p in scalar_space.dofs:
                for d in directions:
                    dofs.append(DotPointEvaluation(reference, p.dof_point(), d, entity=p.entity))

            poly += quolynomial_set_vector(reference.tdim, reference.tdim, order)

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["vector Q", "vQ"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 0
    continuity = "C0"


class Nedelec(CiarletElement):
    """Nedelec Hcurl finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        poly: typing.List[FunctionInput] = []
        poly += quolynomial_set_vector(reference.tdim, reference.tdim, order - 1)
        poly += Hcurl_quolynomials(reference.tdim, reference.tdim, order)

        dofs: ListOfFunctionals = make_integral_moment_dofs(
            reference,
            edges=(TangentIntegralMoment, Q, order - 1, {"variant": variant}),
            faces=(IntegralMoment, RaviartThomas, order - 1, "covariant", {"variant": variant}),
            volumes=(IntegralMoment, RaviartThomas, order - 1, "covariant", {"variant": variant}),
        )

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["NCE", "RTCE", "Qcurl", "Nedelec", "Ncurl"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 1
    continuity = "H(curl)"


class RaviartThomas(CiarletElement):
    """Raviart-Thomas Hdiv finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        poly: typing.List[FunctionInput] = []
        poly += quolynomial_set_vector(reference.tdim, reference.tdim, order - 1)
        poly += Hdiv_quolynomials(reference.tdim, reference.tdim, order)

        dofs: ListOfFunctionals = make_integral_moment_dofs(
            reference,
            facets=(NormalIntegralMoment, Q, order - 1, {"variant": variant}),
            cells=(IntegralMoment, Nedelec, order - 1, "contravariant", {"variant": variant}),
        )

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["NCF", "RTCF", "Qdiv"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 1
    continuity = "H(div)"



"""Taylor element on an interval, triangle or tetrahedron."""

import typing
from itertools import product

from ..finite_element import CiarletElement
from ..functionals import DerivativePointEvaluation, IntegralMoment, ListOfFunctionals
from ..functions import FunctionInput
from ..moments import make_integral_moment_dofs
from ..polynomials import polynomial_set_1d
from ..references import Reference
from .lagrange import Lagrange


class Taylor(CiarletElement):
    """Taylor finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        dofs: ListOfFunctionals = make_integral_moment_dofs(
            reference,
            cells=(IntegralMoment, Lagrange, 0, {"variant": "equispaced"}),
        )
        for i in product(range(order + 1), repeat=reference.tdim):
            if 1 <= sum(i) <= order:
                dofs.append(DerivativePointEvaluation(
                    reference, reference.midpoint(), i, entity=(reference.tdim, 0)))

        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_1d(reference.tdim, order)

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)

    names = ["Taylor", "discontinuous Taylor"]
    references = ["interval", "triangle", "tetrahedron"]
    min_order = 0
    continuity = "L2"



"""Conforming Crouzeix-Raviart elements on simplices.

This element's definition appears in https://doi.org/10.1051/m2an/197307R300331
(Crouzeix, Raviart, 1973)
"""

import typing

import sympy

from ..finite_element import CiarletElement
from ..functionals import ListOfFunctionals, PointEvaluation
from ..functions import FunctionInput
from ..polynomials import polynomial_set_1d
from ..references import Reference
from ..symbols import x


class ConformingCrouzeixRaviart(CiarletElement):
    """Conforming Crouzeix-Raviart finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        if reference.vertices != reference.reference_vertices:
            raise NotImplementedError()
        assert reference.name == "triangle"

        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_1d(reference.tdim, order)

        poly += [
            x[0] ** i * x[1] ** (order - i) * (x[0] + x[1])
            for i in range(1, order)
        ]

        dofs: ListOfFunctionals = []
        for i, v in enumerate(reference.vertices):
            dofs.append(PointEvaluation(reference, v, entity=(0, i)))
        if order >= 2:
            for i, edge in enumerate(reference.edges):
                for p in range(1, order):
                    v = tuple(sympy.Rational((order - p) * a + p * b, order) for a, b in zip(
                        reference.vertices[edge[0]], reference.vertices[edge[1]]))
                    dofs.append(PointEvaluation(reference, v, entity=(1, i)))
            for i in range(1, order):
                for j in range(1, order + 1 - i):
                    point = (
                        sympy.Rational(3 * i - 1, 3 * order),
                        sympy.Rational(3 * j - 1, 3 * order)
                    )
                    dofs.append(PointEvaluation(reference, point, entity=(2, 0)))

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)

    names = ["conforming Crouzeix-Raviart", "conforming CR"]
    references = ["triangle"]
    min_order = 1
    continuity = "L2"



"""Hellan-Herrmann-Johnson elements on simplices.

This element's definition appears in https://arxiv.org/abs/1909.09687
(Arnold, Walker, 2020)
"""

import typing

from ..finite_element import CiarletElement
from ..functionals import IntegralMoment, ListOfFunctionals, NormalInnerProductIntegralMoment
from ..functions import FunctionInput
from ..moments import make_integral_moment_dofs
from ..polynomials import polynomial_set_vector
from ..references import Reference
from .lagrange import Lagrange, SymmetricMatrixLagrange


class HellanHerrmannJohnson(CiarletElement):
    """A Hellan-Herrmann-Johnson element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        if reference.vertices != reference.reference_vertices:
            raise NotImplementedError()
        assert reference.name == "triangle"

        poly: typing.List[FunctionInput] = []
        poly += [((p[0], p[1]), (p[1], p[2]))
                 for p in polynomial_set_vector(reference.tdim, 3, order)]

        dofs: ListOfFunctionals = make_integral_moment_dofs(
            reference,
            facets=(NormalInnerProductIntegralMoment, Lagrange, order,
                    {"variant": variant}),
            cells=(IntegralMoment, SymmetricMatrixLagrange, order - 1,
                   {"variant": variant}),
        )

        self.variant = variant

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim ** 2,
                         (reference.tdim, reference.tdim))

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["Hellan-Herrmann-Johnson", "HHJ"]
    references = ["triangle"]
    min_order = 0
    continuity = "inner H(div)"



"""Lagrange elements on a prism."""

import typing
from itertools import product

import sympy

from ..finite_element import CiarletElement
from ..functionals import DotPointEvaluation, ListOfFunctionals, PointEvaluation
from ..functions import FunctionInput
from ..polynomials import prism_polynomial_set_1d, prism_polynomial_set_vector
from ..quadrature import get_quadrature
from ..references import Reference


class Lagrange(CiarletElement):
    """Lagrange finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        dofs: ListOfFunctionals = []
        if order == 0:
            dofs = [
                PointEvaluation(
                    reference, tuple(
                        sympy.Rational(1, reference.tdim + 1)
                        for i in range(reference.tdim)
                    ),
                    entity=(reference.tdim, 0)
                )
            ]
        else:
            points, _ = get_quadrature(variant, order + 1)

            # Vertices
            for v_n, v in enumerate(reference.vertices):
                dofs.append(PointEvaluation(reference, v, entity=(0, v_n)))
            # Edges
            for e_n in range(reference.sub_entity_count(1)):
                entity = reference.sub_entity(1, e_n)
                for i in range(1, order):
                    dofs.append(
                        PointEvaluation(
                            reference, tuple(o + entity.axes[0][j] * points[i]
                                             for j, o in enumerate(entity.origin)),
                            entity=(1, e_n)))
            # Faces
            for e_n in range(reference.sub_entity_count(2)):
                entity = reference.sub_entity(2, e_n)
                for ii in product(range(1, order), repeat=2):
                    if len(entity.vertices) == 4 or sum(ii) < order:
                        dofs.append(
                            PointEvaluation(
                                reference, tuple(
                                    o + sum(a[j] * points[b] for a, b in zip(entity.axes, ii[::-1]))
                                    for j, o in enumerate(entity.origin)), entity=(2, e_n)))

            # Interior
            for ii in product(range(1, order), repeat=3):
                if ii[0] + ii[1] < order:
                    dofs.append(
                        PointEvaluation(
                            reference, tuple(
                                o + sum(a[j] * points[b] for a, b in zip(reference.axes, ii))
                                for j, o in enumerate(reference.origin)), entity=(3, 0)))

        poly: typing.List[FunctionInput] = []
        poly += prism_polynomial_set_1d(reference.tdim, order)

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["Lagrange", "P"]
    references = ["prism"]
    min_order = 0
    continuity = "C0"


class VectorLagrange(CiarletElement):
    """Vector Lagrange finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        scalar_space = Lagrange(reference, order, variant)
        dofs: ListOfFunctionals = []
        poly: typing.List[FunctionInput] = []
        if reference.tdim == 1:
            for p in scalar_space.dofs:
                dofs.append(PointEvaluation(reference, p.dof_point(), entity=p.entity))

            poly += prism_polynomial_set_1d(reference.tdim, order)
        else:
            directions = [
                tuple(1 if i == j else 0 for j in range(reference.tdim))
                for i in range(reference.tdim)
            ]
            for p in scalar_space.dofs:
                for d in directions:
                    dofs.append(DotPointEvaluation(reference, p.dof_point(), d, entity=p.entity))

            poly += prism_polynomial_set_vector(reference.tdim, reference.tdim, order)

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names: typing.List[str] = []
    references = ["prism"]
    min_order = 0
    continuity = "C0"



"""Serendipity elements on tensor product cells.

This element's definition appears in https://doi.org/10.1007/s10208-011-9087-3
(Arnold, Awanou, 2011)
"""

import typing

from ..finite_element import CiarletElement
from ..functionals import (IntegralMoment, ListOfFunctionals, NormalIntegralMoment, PointEvaluation,
                           TangentIntegralMoment)
from ..functions import FunctionInput
from ..moments import make_integral_moment_dofs
from ..polynomials import (Hcurl_serendipity, Hdiv_serendipity, polynomial_set_1d,
                           polynomial_set_vector, serendipity_set_1d)
from ..references import Reference
from .dpc import DPC, VectorDPC


class Serendipity(CiarletElement):
    """A serendipity element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_1d(reference.tdim, order)
        poly += serendipity_set_1d(reference.tdim, order)

        dofs: ListOfFunctionals = []
        for v_n, p in enumerate(reference.vertices):
            dofs.append(PointEvaluation(reference, p, entity=(0, v_n)))
        dofs += make_integral_moment_dofs(
            reference,
            edges=(IntegralMoment, DPC, order - 2, {"variant": variant}),
            faces=(IntegralMoment, DPC, order - 4, {"variant": variant}),
            volumes=(IntegralMoment, DPC, order - 6, {"variant": variant}),
        )

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["serendipity", "S"]
    references = ["interval", "quadrilateral", "hexahedron"]
    min_order = 1
    continuity = "C0"


class SerendipityCurl(CiarletElement):
    """A serendipity Hcurl element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_vector(reference.tdim, reference.tdim, order)
        poly += Hcurl_serendipity(reference.tdim, reference.tdim, order)

        dofs: ListOfFunctionals = make_integral_moment_dofs(
            reference,
            edges=(TangentIntegralMoment, DPC, order, {"variant": variant}),
            faces=(IntegralMoment, VectorDPC, order - 2, "covariant",
                   {"variant": variant}),
            volumes=(IntegralMoment, VectorDPC, order - 4, "covariant",
                     {"variant": variant}),
        )

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["serendipity Hcurl", "Scurl", "BDMCE", "AAE"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 1
    continuity = "H(curl)"


class SerendipityDiv(CiarletElement):
    """A serendipity Hdiv element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_vector(reference.tdim, reference.tdim, order)
        poly += Hdiv_serendipity(reference.tdim, reference.tdim, order)

        dofs: ListOfFunctionals = make_integral_moment_dofs(
            reference,
            facets=(NormalIntegralMoment, DPC, order, {"variant": variant}),
            cells=(IntegralMoment, VectorDPC, order - 2, "contravariant",
                   {"variant": variant}),
        )

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["serendipity Hdiv", "Sdiv", "BDMCF", "AAF"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 1
    continuity = "H(div)"



"""Utility functions."""

import typing

import sympy

from .functions import ScalarFunction


def allequal(a: typing.Any, b: typing.Any) -> bool:
    """Test if two items that may be nested lists/tuples are equal.

    Args:
        a: The first item
        b: The second item

    Returns:
        a == b?
    """
    if isinstance(a, (tuple, list)) and isinstance(b, (tuple, list)):
        for i, j in zip(a, b):
            if not allequal(i, j):
                return False
        return True
    if isinstance(a, sympy.core.expr.Expr):
        a = ScalarFunction(a)
    if isinstance(b, sympy.core.expr.Expr):
        a = ScalarFunction(b)
    return a == b



"""DPC elements on tensor product cells."""

import typing
from itertools import product

import sympy

from ..finite_element import CiarletElement
from ..functionals import DotPointEvaluation, ListOfFunctionals, PointEvaluation
from ..functions import FunctionInput
from ..polynomials import polynomial_set_1d, polynomial_set_vector
from ..references import Reference
from .lagrange import Lagrange


class DPC(CiarletElement):
    """A dPc element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        if reference.name == "interval":
            points = [d.dof_point() for d in Lagrange(reference, order, variant).dofs]
        elif order == 0:
            points = [tuple(sympy.Rational(1, 2) for _ in range(reference.tdim))]
        else:
            points = [
                tuple(sympy.Rational(j, order) for j in i[::-1])
                for i in product(range(order + 1), repeat=reference.tdim)
                if sum(i) <= order
            ]

        dofs: ListOfFunctionals = [
            PointEvaluation(reference, d, entity=(reference.tdim, 0)) for d in points]

        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_1d(reference.tdim, order)
        super().__init__(
            reference, order, poly, dofs, reference.tdim, 1
        )
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["dPc"]
    references = ["interval", "quadrilateral", "hexahedron"]
    min_order = 0
    continuity = "L2"


class VectorDPC(CiarletElement):
    """Vector dPc finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        scalar_space = DPC(reference, order, variant)
        dofs: ListOfFunctionals = []
        if reference.tdim == 1:
            directions: typing.List[typing.Tuple[int, ...]] = [(1, )]
        else:
            directions = [
                tuple(1 if i == j else 0 for j in range(reference.tdim))
                for i in range(reference.tdim)
            ]
        for p in scalar_space.dofs:
            for d in directions:
                dofs.append(DotPointEvaluation(reference, p.dof_point(), d, entity=p.entity))

        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_vector(reference.tdim, reference.tdim, order)

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["vector dPc"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 0
    continuity = "L2"



"""Test tensor product factorisations."""

import pytest
import sympy

import symfem
from symfem import create_element
from symfem.utils import allequal

from .utils import test_elements


def make_lattice(cell, N=3):
    if cell == "interval":
        return [[sympy.Rational(i, N)] for i in range(N + 1)]
    if cell == "triangle":
        return [[sympy.Rational(i, N), sympy.Rational(j, N)]
                for i in range(N + 1) for j in range(N + 1 - i)]
    if cell == "tetrahedron":
        return [[sympy.Rational(i, N), sympy.Rational(j, N), sympy.Rational(k, N)]
                for i in range(N + 1) for j in range(N + 1 - i) for k in range(N + 1 - i - j)]
    if cell == "quadrilateral":
        return [[sympy.Rational(i, N), sympy.Rational(j, N)]
                for i in range(N + 1) for j in range(N + 1)]
    if cell == "hexahedron":
        return [[sympy.Rational(i, N), sympy.Rational(j, N), sympy.Rational(k, N)]
                for i in range(N + 1) for j in range(N + 1) for k in range(N + 1)]
    if cell == "prism":
        return [[sympy.Rational(i, N), sympy.Rational(j, N), sympy.Rational(k, N)]
                for i in range(N + 1) for j in range(N + 1 - i) for k in range(N + 1)]
    if cell == "pyramid":
        return [[sympy.Rational(i, N), sympy.Rational(j, N), sympy.Rational(k, N)]
                for i in range(N + 1) for j in range(N + 1) for k in range(N + 1 - max(i, j))]


@pytest.mark.parametrize(
    ("cell_type", "element_type", "order", "kwargs"),
    [[reference, element, order, kwargs]
     for reference, i in test_elements.items() for element, j in i.items()
     for kwargs, k in j for order in k])
def test_element(
    elements_to_test, cells_to_test, cell_type, element_type, order, kwargs,
    speed
):
    """Run tests for each element."""
    if elements_to_test != "ALL" and element_type not in elements_to_test:
        pytest.skip()
    if cells_to_test != "ALL" and cell_type not in cells_to_test:
        pytest.skip()
    if speed == "fast":
        if order > 2:
            pytest.skip()
        if order == 2 and cell_type in ["tetrahedron", "hexahedron", "prism", "pyramid"]:
            pytest.skip()

    element = create_element(cell_type, element_type, order, **kwargs)
    try:
        factorised_basis = element._get_basis_functions_tensor()
    except symfem.finite_element.NoTensorProduct:
        pytest.skip("This element does not have a tensor product representation.")
    basis = element.get_basis_functions()

    for i, j in zip(basis, factorised_basis):
        assert allequal(i, j)



"""Arnold-Winther elements on simplices.

Thse elements definitions appear in https://doi.org/10.1007/s002110100348
(Arnold, Winther, 2002) [conforming] and https://doi.org/10.1142/S0218202503002507
(Arnold, Winther, 2003) [nonconforming]
"""

import typing

import sympy

from ..finite_element import CiarletElement
from ..functionals import (InnerProductIntegralMoment, IntegralMoment, ListOfFunctionals,
                           PointInnerProduct, VecIntegralMoment)
from ..functions import FunctionInput
from ..polynomials import polynomial_set_vector
from ..references import Reference
from ..symbols import x
from .lagrange import Lagrange


class ArnoldWinther(CiarletElement):
    """An Arnold-Winther element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        assert reference.name == "triangle"
        self.variant = variant
        poly: typing.List[FunctionInput] = []
        poly += [
            ((p[0], p[1]), (p[1], p[2]))
            for p in polynomial_set_vector(reference.tdim, 3, order - 1)]
        poly += [(((order - k + 1) * (order - k + 2) * x[0] ** k * x[1] ** (order - k),
                   -k * (order - k + 2) * x[0] ** (k - 1) * x[1] ** (order - k + 1)),
                  (-k * (order - k + 2) * x[0] ** (k - 1) * x[1] ** (order - k + 1),
                   -k * (k - 1) * x[0] ** (k - 2) * x[1] ** (order - k + 2)))
                 for k in range(order + 1)]
        poly += [((0, x[0] ** order), (x[0] ** order, -order * x[0] ** (order - 1) * x[1])),
                 ((0, 0), (0, x[0] ** order))]

        dofs: ListOfFunctionals = []
        for v_n, v in enumerate(reference.vertices):
            for d in [[(1, 0), (1, 0)],
                      [(1, 0), (0, 1)],
                      [(0, 1), (0, 1)]]:
                dofs.append(PointInnerProduct(reference, v, d[0], d[1], entity=(0, v_n),
                                              mapping="double_contravariant"))
        for e_n in range(reference.sub_entity_count(1)):
            sub_ref = reference.sub_entity(1, e_n)
            sub_e = Lagrange(sub_ref.default_reference(), order - 2, variant)
            for dof_n, dof in enumerate(sub_e.dofs):
                p = sub_e.get_basis_function(dof_n).get_function()
                for component in [sub_ref.normal(), sub_ref.tangent()]:
                    InnerProductIntegralMoment(
                        reference, sub_ref, p, component, sub_ref.normal(), dof,
                        entity=(1, e_n), mapping="double_contravariant")
                    dofs.append(
                        InnerProductIntegralMoment(
                            reference, sub_ref, p, component, sub_ref.normal(), dof,
                            entity=(1, e_n), mapping="double_contravariant"))
        sub_e = Lagrange(reference, order - 3, variant)
        for dof_n, dof in enumerate(sub_e.dofs):
            p = sub_e.get_basis_function(dof_n).get_function()
            for component22 in [((1, 0), (0, 0)), ((0, 1), (0, 0)), ((0, 0), (0, 1))]:
                dofs.append(VecIntegralMoment(
                    reference, reference, p, component22, dof, entity=(2, 0)))

        if order >= 4:
            sub_e = Lagrange(reference, order - 4, variant)
            for p, dof in zip(sub_e.get_basis_functions(), sub_e.dofs):
                if sympy.Poly(p.as_sympy(), x[:2]).degree() != order - 4:
                    continue
                f = p * x[0] ** 2 * x[1] ** 2 * (1 - x[0] - x[1]) ** 2
                J = tuple(tuple(f.diff(x[i]).diff(x[j]) for j in range(2))
                          for i in range(2))
                dofs.append(IntegralMoment(reference, reference, J, dof, entity=(2, 0)))

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim ** 2,
                         (reference.tdim, reference.tdim))

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["Arnold-Winther", "AW", "conforming Arnold-Winther"]
    references = ["triangle"]
    min_order = 3
    continuity = "integral inner H(div)"


class NonConformingArnoldWinther(CiarletElement):
    """A nonconforming Arnold-Winther element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        assert reference.name == "triangle"
        self.variant = variant
        poly: typing.List[FunctionInput] = []
        poly += [
            ((p[0], p[1]), (p[1], p[2]))
            for p in polynomial_set_vector(reference.tdim, 3, order - 1)]

        poly += [
            ((0, x[1] ** 2), (x[1] ** 2, -2 * x[1] ** 2)),
            ((-2 * x[0] ** 2, x[0] ** 2), (x[0] ** 2, 0)),
            ((-2 * x[0] * x[1], x[0] * x[1]), (x[0] * x[1], 0)),
            ((x[0] * (x[0] - x[1]), 0), (0, 0)),
            ((x[0] ** 2, 0), (0, x[0] * x[1])),
            ((x[0] ** 2, 0), (0, x[1] ** 2))
        ]

        dofs: ListOfFunctionals = []
        for e_n in range(reference.sub_entity_count(1)):
            sub_ref = reference.sub_entity(1, e_n)
            sub_e = Lagrange(sub_ref.default_reference(), 1, variant)
            for dof_n, dof in enumerate(sub_e.dofs):
                p = sub_e.get_basis_function(dof_n).get_function()
                for component in [sub_ref.normal(), sub_ref.tangent()]:
                    dofs.append(
                        InnerProductIntegralMoment(
                            reference, sub_ref, p, component, sub_ref.normal(), dof,
                            entity=(1, e_n), mapping="double_contravariant"))
        sub_e = Lagrange(reference, 0, variant)
        for dof_n, dof in enumerate(sub_e.dofs):
            p = sub_e.get_basis_function(dof_n).get_function()
            for component22 in [((1, 0), (0, 0)), ((0, 1), (0, 0)), ((0, 0), (0, 1))]:
                dofs.append(VecIntegralMoment(
                    reference, reference, p, component22, dof, entity=(2, 0)))

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim ** 2,
                         (reference.tdim, reference.tdim))

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["nonconforming Arnold-Winther", "nonconforming AW"]
    references = ["triangle"]
    min_order = 2
    max_order = 2
    continuity = "integral inner H(div)"



"""Brezzi-Douglas-Fortin-Marini elements.

This element's definition appears in https://doi.org/10.1051/m2an/1987210405811
(Brezzi, Douglas, Fortin, Marini, 1987)
"""

import typing

from ..finite_element import CiarletElement
from ..functionals import IntegralMoment, ListOfFunctionals, NormalIntegralMoment
from ..functions import FunctionInput
from ..moments import make_integral_moment_dofs
from ..polynomials import polynomial_set_vector
from ..references import Reference
from ..symbols import x
from .dpc import DPC, VectorDPC
from .lagrange import Lagrange, VectorLagrange


def bdfm_polyset(reference: Reference, order: int) -> typing.List[FunctionInput]:
    """Create the polynomial basis for a BDFM element.

    Args:
        reference: The reference cell
        order: The polynomial order

    Returns:
        The polynomial basis
    """
    dim = reference.tdim
    pset: typing.List[FunctionInput] = []
    pset += polynomial_set_vector(dim, dim, order - 1)
    if reference.name == "quadrilateral":
        for i in range(1, order + 1):
            j = order - i
            pset.append((x[0] ** i * x[1] ** j, 0))
            pset.append((0, x[1] ** i * x[0] ** j))
    elif reference.name == "triangle":
        for i in range(order):
            p = x[0] ** i * x[1] ** (order - 1 - i)
            pset.append((x[0] * p, x[1] * p))
    elif reference.name == "hexahedron":
        for i in range(1, order + 1):
            for j in range(order + 1 - i):
                k = order - i - j
                pset.append((x[0] ** i * x[1] ** j * x[2] ** k, 0, 0))
                pset.append((0, x[1] ** i * x[0] ** j * x[2] ** k, 0))
                pset.append((0, 0, x[2] ** i * x[0] ** j * x[1] ** k))
    elif reference.name == "tetrahedron":
        for i in range(order):
            for j in range(order - i):
                p = x[0] ** i * x[1] ** j * x[2] ** (order - 1 - i - j)
                pset.append((x[0] * p, x[1] * p, x[2] * p))
    return pset


class BDFM(CiarletElement):
    """Brezzi-Douglas-Fortin-Marini Hdiv finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        poly = bdfm_polyset(reference, order)

        dofs: ListOfFunctionals = []
        if reference.name in ["triangle", "tetrahedron"]:
            dofs = make_integral_moment_dofs(
                reference,
                facets=(NormalIntegralMoment, Lagrange, order - 1, {"variant": variant}),
                cells=(IntegralMoment, VectorLagrange, order - 2, {"variant": variant}),
            )
        else:
            dofs = make_integral_moment_dofs(
                reference,
                facets=(NormalIntegralMoment, DPC, order - 1, {"variant": variant}),
                cells=(IntegralMoment, VectorDPC, order - 2, {"variant": variant}),
            )
        self.variant = variant

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["Brezzi-Douglas-Fortin-Marini", "BDFM"]
    references = ["triangle", "quadrilateral", "hexahedron", "tetrahedron"]
    min_order = 1
    continuity = "H(div)"
